"""Configuration class for linear constraints."""

from __future__ import annotations

from typing import TYPE_CHECKING, Self

import numpy as np
from pydantic import ConfigDict, model_validator

from ropt.config.utils import (
    ImmutableBaseModel,
    broadcast_1d_array,
    immutable_array,
)
from ropt.config.validated_types import Array1D, Array2D  # noqa: TC001

if TYPE_CHECKING:
    from ropt.config.enopt import VariablesConfig
    from ropt.transforms import OptModelTransforms


class LinearConstraintsConfig(ImmutableBaseModel):
    r"""Configuration class for linear constraints.

    This class, `LinearConstraintsConfig`, defines linear constraints used in an
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.

    Linear constraints are defined by a set of linear equations involving the
    optimization variables. These equations can represent equality or inequality
    constraints. The `coefficients` field is a 2D `numpy` array where each row
    represents a constraint, and each column corresponds to a variable.

    The `lower_bounds` and `upper_bounds` fields specify the bounds on the
    right-hand side of each constraint equation. These fields are converted and
    broadcasted to `numpy` arrays with a length equal to the number of
    constraint equations.

    Less-than and greater-than inequality constraints can be specified by
    setting the lower bounds to $-\infty$, or the upper bounds to $+\infty$,
    respectively. Equality constraints are specified by setting the lower bounds
    equal to the upper bounds.

    Attributes:
        coefficients: Matrix of coefficients for the linear constraints.
        lower_bounds: Lower bounds for the right-hand side of the constraint equations.
        upper_bounds: Upper bounds for the right-hand side of the constraint equations.

    Note: Linear transformation of variables.
        The set of linear constraints can be represented by a matrix equation:
        $\mathbf{A} \mathbf{x} = \mathbf{b}$.

        When linearly transforming variables to the optimizer domain, the
        coefficients ($\mathbf{A}$) and right-hand-side values ($\mathbf{b}$)
        must be converted to remain valid. If the linear transformation of the
        variables to the optimizer domain is given by:

        $$ \hat{\mathbf{x}} = \mathbf{S} \mathbf{x} + \mathbf{o}$$

        then the coefficients and right-hand-side values must be transformed as
        follows:

        $$ \begin{align}
            \hat{\mathbf{A}} &= \mathbf{A} \mathbf{S}^{-1} \\ \hat{\mathbf{b}}
            &= \mathbf{b} + \mathbf{A}\mathbf{S}^{-1}\mathbf{o}
        \end{align}$$
    """

    coefficients: Array2D
    lower_bounds: Array1D
    upper_bounds: Array1D

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _broadcast_and_check(self) -> Self:
        size = 0 if self.coefficients is None else self.coefficients.shape[0]
        lower_bounds = broadcast_1d_array(self.lower_bounds, "lower_bounds", size)
        upper_bounds = broadcast_1d_array(self.upper_bounds, "upper_bounds", size)
        if np.any(lower_bounds > upper_bounds):
            msg = "The lower bounds are larger than the upper bounds."
            raise ValueError(msg)

        self._mutable()
        self.lower_bounds = immutable_array(lower_bounds)
        self.upper_bounds = immutable_array(upper_bounds)
        self._immutable()
        return self

    def apply_transformation(
        self, variables: VariablesConfig, transforms: OptModelTransforms | None
    ) -> LinearConstraintsConfig:
        variable_count = variables.initial_values.size
        if (
            self.coefficients.shape[0] > 0
            and self.coefficients.shape[1] != variable_count
        ):
            msg = f"the coefficients matrix should have {variable_count} columns"
            raise ValueError(msg)

        if transforms is not None and transforms.variables is not None:
            coefficients, lower_bounds, upper_bounds = (
                transforms.variables.linear_constraints_to_optimizer(
                    self.coefficients, self.lower_bounds, self.upper_bounds
                )
            )
            values = self.model_dump(round_trip=True)
            values.update(
                coefficients=coefficients,
                lower_bounds=lower_bounds,
                upper_bounds=upper_bounds,
            )
            return LinearConstraintsConfig.model_construct(**values)

        return self



"""This module defines the abstract base class for samplers."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from ropt.plugins.base import Plugin

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig


class Sampler(ABC):
    """Abstract base class for sampler classes.

    `ropt` employs plugins to implement samplers that are called during an
    optimization plan to generate perturbed variable vectors. Samplers should
    derive from the `Sampler` base class, which specifies the requirements for
    the class constructor (`__init__`) and also includes a `generate_samples`
    method used to generate samples used to create perturbed values.
    """

    def __init__(  # noqa: B027
        self,
        enopt_config: EnOptConfig,
        sampler_index: int,
        mask: NDArray[np.bool_] | None,
        rng: Generator,
    ) -> None:
        """Initialize the sampler object.

        The `samplers` field in the
        [`enopt_config`][ropt.config.enopt.EnOptConfig] configuration used by
        the optimization is a tuple of sampler configurations (see
        [`SamplerConfig`][ropt.config.enopt.SamplerConfig]). The `sampler_index`
        field is used to identify the configuration to use to initialize this
        sampler.

        The sampler may be used for a subset of the variables. The boolean
        `mask` array indicates the variables that are handled by this sampler.

        Args:
            enopt_config:  The configuration of the optimizer.
            sampler_index: The index of the sampler to use.
            mask:          The mask of the variables to sample.
            rng:           A random generator object for use by stochastic samplers.
        """

    @abstractmethod
    def generate_samples(self) -> NDArray[np.float64]:
        """Return an array containing sampled values.

        The result should be a three-dimensional array of perturbation values.
        The variable values are stored along the last axis, for each realization
        and perturbation. The first axis indexes the realization, and the second
        axis indexes the perturbation.

        If the `shared` flag is set in the
        [`SamplerConfig`][ropt.config.enopt.SamplerConfig] configuration, the
        first dimension should have a length equal to one, since all
        realizations will use the same set of perturbations.

        The sampler may handle only a subset of the variables, as specified by
        the `variable_indices` argument of the constructor. In this case, only
        the corresponding values along the variables axis (the last axis) should
        be set, while other values should be zero.

        Note: Sample scaling
            Samples will be multiplied by the values given by the
            `perturbation_magnitudes` field in the
            [`gradients`][ropt.config.enopt.GradientConfig] section of the
            optimizer configuration. It makes therefore sense to generate
            samples that have an order of magnitude around one. For instance, by
            generating them on a `[-1, 1]` range, or with a unit standard
            deviation.

        Returns:
            The sampled values.
        """


class SamplerPlugin(Plugin):
    """Abtract base class for sampler plugins."""

    @classmethod
    @abstractmethod
    def create(
        cls,
        enopt_config: EnOptConfig,
        sampler_index: int,
        mask: NDArray[np.bool_] | None,
        rng: Generator,
    ) -> Sampler:
        """Create a sampler.

        Args:
            enopt_config:  The configuration of the optimizer.
            sampler_index: The index of the sampler to use.
            mask:          The indices of the variables to sample.
            rng:           A random generator object for use by stochastic samplers.
        """



"""Exceptions raised within the `ropt` library."""

from .enums import OptimizerExitCode


class ConfigError(Exception):
    """Raised when an configuration error occurs."""


class OptimizationAborted(Exception):  # noqa: N818
    """Raised when an optimization is aborted.

    When constructing the exception object an exit code must be passed that
    indicates the reason for aborting (see
    [`OptimizerExitCode`][ropt.enums.OptimizerExitCode]).
    """

    def __init__(self, exit_code: OptimizerExitCode) -> None:
        """Initialize an exception that aborts the optimization.

        Args:
            exit_code: The exit code indicating the reason for the abort.
        """
        self.exit_code = exit_code
        super().__init__()


class PlanAborted(Exception):  # noqa: N818
    """Raised when an optimization plan is aborted.

    Steps within a plan may abort a plan by setting its `aborted` attribute. Any
    attemps to run steps after this, will raise this exception.
    """



import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig
from ropt.plugins.function_estimator.base import FunctionEstimator


def _calculate_estimated_functions(  # noqa: PLR0913
    config: EnOptConfig,
    function_estimators: list[FunctionEstimator],
    estimator_indices: NDArray[np.intc] | None,
    functions: NDArray[np.float64],
    realization_weights: NDArray[np.float64] | None,
    failed_realizations: NDArray[np.bool_],
) -> NDArray[np.float64]:
    result: NDArray[np.float64] = np.empty(functions.shape[-1], dtype=np.float64)

    if estimator_indices is None:
        estimator_indices = np.zeros(functions.shape[1], dtype=np.intc)

    for estimator_idx, estimator in enumerate(function_estimators):
        mask = estimator_indices == estimator_idx
        for idx in np.where(mask)[0]:
            weights = (
                config.realizations.weights
                if realization_weights is None
                else realization_weights[idx, ...]
            )
            weights = np.where(failed_realizations, 0.0, weights)
            weights /= weights.sum()
            result[idx] = estimator.calculate_function(functions[..., idx], weights)

    return result


def _calculate_estimated_objectives(
    config: EnOptConfig,
    function_estimators: list[FunctionEstimator],
    functions: NDArray[np.float64],
    realization_weights: NDArray[np.float64] | None,
    failed_realizations: NDArray[np.bool_],
) -> NDArray[np.float64]:
    return _calculate_estimated_functions(
        config,
        function_estimators,
        config.objectives.function_estimators,
        functions,
        realization_weights,
        failed_realizations,
    )


def _calculate_estimated_constraints(
    config: EnOptConfig,
    function_estimators: list[FunctionEstimator],
    constraints: NDArray[np.float64] | None,
    realization_weights: NDArray[np.float64] | None,
    failed_realizations: NDArray[np.bool_],
) -> NDArray[np.float64] | None:
    if constraints is None:
        return None
    assert config.nonlinear_constraints is not None
    return _calculate_estimated_functions(
        config,
        function_estimators,
        config.nonlinear_constraints.function_estimators,
        constraints,
        realization_weights,
        failed_realizations,
    )



"""This module implements the default function estimator plugin."""

from typing import Final

import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig
from ropt.enums import OptimizerExitCode
from ropt.exceptions import ConfigError, OptimizationAborted

from .base import FunctionEstimator, FunctionEstimatorPlugin

_MIN_STDDEV_REALIZATIONS: Final = 2


class DefaultFunctionEstimator(FunctionEstimator):
    """The default function estimator plugin.

    This plugin currently implements two methods:

    `mean`:
    :  Calculate the combined functions as a weighted mean of the function
       values of each realization. Gradients are accordingly calculated as
       a weighted sum.

    `stddev`:
    :  Calculate the combined functions as the standard deviation of function
       values of each realization. Gradients are calculated accordingly using
       the chain rule. The sign of the result is adjusted such that the standard
       deviation is always minimized.
    """

    def __init__(self, enopt_config: EnOptConfig, estimator_index: int) -> None:
        """Initialize the function estimator object.

        See the
        [ropt.plugins.function_estimator.base.FunctionEstimator][]
        abstract base class.

        # noqa
        """
        self._enopt_config = enopt_config
        self._estimator_config = enopt_config.function_estimators[estimator_index]
        _, _, self._method = self._estimator_config.method.lower().rpartition("/")
        if self._method == "default":
            self._method = "mean"
        if self._method == "stddev" and self._enopt_config.gradient.merge_realizations:
            msg = (
                "The stddev estimator does not support merging "
                "realizations in the gradient."
            )
            raise ConfigError(msg)

    def calculate_function(
        self, functions: NDArray[np.float64], weights: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Calculate a function from function values for each realization.

        See the
        [ropt.plugins.function_estimator.base.FunctionEstimator][]
        abstract base class.

        # noqa
        """
        estimator_method = self._method
        if estimator_method == "mean":
            return self._calculate_function_mean(functions, weights)
        if estimator_method == "stddev":
            return self._calculate_function_stddev(functions, weights)
        msg = f"Function estimator method not supported: {estimator_method}"
        raise ConfigError(msg)

    def calculate_gradient(
        self,
        functions: NDArray[np.float64],
        gradient: NDArray[np.float64],
        weights: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Calculate a gradient from gradients of the realizations.

        See the
        [ropt.plugins.function_estimator.base.FunctionEstimator][]
        abstract base class.

        # noqa
        """
        estimator_method = self._method
        if estimator_method == "mean":
            return self._calculate_gradient_mean(functions, gradient, weights)
        if estimator_method == "stddev":
            return self._calculate_gradient_stddev(functions, gradient, weights)
        msg = f"Function estimator method not supported: {estimator_method}"
        raise ConfigError(msg)

    @staticmethod
    def _calculate_function_mean(
        functions: NDArray[np.float64], weights: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        functions = np.nan_to_num(functions)
        return np.dot(functions, weights)  # type: ignore[no-any-return]

    def _calculate_gradient_mean(
        self,
        _: NDArray[np.float64],
        gradient: NDArray[np.float64],
        weights: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        if self._enopt_config.gradient.merge_realizations:
            return gradient
        return np.dot(gradient, weights)  # type: ignore[no-any-return]

    def _calculate_function_stddev(
        self, functions: NDArray[np.float64], weights: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        if np.count_nonzero(weights) < _MIN_STDDEV_REALIZATIONS:
            raise OptimizationAborted(exit_code=OptimizerExitCode.TOO_FEW_REALIZATIONS)
        functions = np.nan_to_num(functions)
        *_, stddev = self._mean_stddev(functions, weights)
        return stddev

    def _calculate_gradient_stddev(
        self,
        functions: NDArray[np.float64],
        gradient: NDArray[np.float64],
        weights: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        if np.count_nonzero(weights) < _MIN_STDDEV_REALIZATIONS:
            raise OptimizationAborted(exit_code=OptimizerExitCode.TOO_FEW_REALIZATIONS)
        functions = np.nan_to_num(functions)
        norm, mean, stddev = self._mean_stddev(functions, weights)
        mean_gradient = np.dot(gradient, weights)
        return (
            np.zeros(mean_gradient.shape, dtype=np.float64)
            if np.allclose(np.abs(stddev), 0.0)
            else (
                (norm / stddev)
                * (np.dot(gradient, functions * weights) - mean * mean_gradient)
            )
        )

    def _mean_stddev(
        self,
        functions: NDArray[np.float64],
        weights: NDArray[np.float64],
    ) -> tuple[float, NDArray[np.float64], NDArray[np.float64]]:
        norm = float(np.count_nonzero(weights > 0))
        norm = norm / (norm - 1)
        mean = np.dot(functions, weights)
        stddev = np.sqrt(
            norm * np.dot((functions - mean[..., np.newaxis]) ** 2, weights)
        )
        return norm, mean, stddev


class DefaultFunctionEstimatorPlugin(FunctionEstimatorPlugin):
    """Default filter estimator plugin class."""

    @classmethod
    def create(
        cls, enopt_config: EnOptConfig, estimator_index: int
    ) -> DefaultFunctionEstimator:
        """Initialize the realization filter plugin.

        See the [ropt.plugins.function_estimator.base.FunctionEstimatorPlugin][]
        abstract base class.

        # noqa
        """
        return DefaultFunctionEstimator(enopt_config, estimator_index)

    @classmethod
    def is_supported(cls, method: str) -> bool:
        """Check if a method is supported.

        See the [ropt.plugins.base.Plugin][] abstract base class.

        # noqa
        """
        return method.lower() in {"default", "mean", "stddev"}



"""Configuration class for objective functions."""

from __future__ import annotations

from typing import Self

import numpy as np
from pydantic import ConfigDict, model_validator

from ropt.config.utils import ImmutableBaseModel, normalize
from ropt.config.validated_types import (  # noqa: TC001
    Array1D,
    Array1DInt,
)


class ObjectiveFunctionsConfig(ImmutableBaseModel):
    """Configuration class for objective functions.

    This class, `ObjectiveFunctionsConfig`, defines the configuration for
    objective functions used in an
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.

    `ropt` supports multi-objective optimization. Multiple objectives are
    combined into a single value by summing them after weighting. The `weights`
    field, a `numpy` array, determines the weight of each objective function.
    The length of this array defines the number of objective functions. The
    weights are automatically normalized to sum to 1 (e.g., `[1, 1]` becomes
    `[0.5, 0.5]`).

    Objective functions can be processed by realization filters and function
    estimators. The `realization_filters` and `function_estimators` fields
    contain indices that refer to the corresponding configuration objects in the
    parent [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.

    Attributes:
        weights:             Weights for the objective functions (default: 1.0).
        realization_filters: Optional indices of realization filters.
        function_estimators: Optional indices of function estimators.
    """

    weights: Array1D = np.array(1.0)
    realization_filters: Array1DInt | None = None
    function_estimators: Array1DInt | None = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _broadcast_and_normalize(self) -> Self:
        self._mutable()
        self.weights = normalize(self.weights)
        self._immutable()
        return self



"""This module defines the abstract base classes for function estimators."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from ropt.plugins.base import Plugin

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig


class FunctionEstimator(ABC):
    """Abstract base class for function estimators."""

    def __init__(self, enopt_config: EnOptConfig, estimator_index: int) -> None:  # noqa: B027
        """Initialize the function estimator object.

        Args:
            enopt_config:    The configuration of the optimizer.
            estimator_index: The index of the estimator to use.
        """

    @abstractmethod
    def calculate_function(
        self, functions: NDArray[np.float64], weights: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Combine functions from realizations into an expected function.

        Tip: Calculation from merged realizations
           Normally the gradient is calculated for each realization separately
           and then combined into an overall gradient with `calculate_gradient`
           method. The `merge_realizations` flag in the ensemble optimizer
           configuration directs the optimizer to calculate the overall gradient
           from all realizations directly. This yields a reasonable estimation
           if the function estimator is an averaging operation, but may not be
           appropriate in other cases.

           At initialization, the `merge_realizations` flag should be checked,
           and if necessary a `ConfigError` with an appropriate message should
           be raised.

        Args:
            functions: The functions for each realization.
            weights:   The weight of each realization.

        Returns:
            The expected function values.
        """

    @abstractmethod
    def calculate_gradient(
        self,
        functions: NDArray[np.float64],
        gradient: NDArray[np.float64],
        weights: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Combine gradients from realizations into an expected gradient.

        Args:
            functions: The functions for each realization.
            gradient:  The gradient for each realization.
            weights:   The weight of each realization.

        Returns:
            The expected gradients.
        """


class FunctionEstimatorPlugin(Plugin):
    """The function estimator plugin base class."""

    @classmethod
    @abstractmethod
    def create(
        cls, enopt_config: EnOptConfig, estimator_index: int
    ) -> FunctionEstimator:
        """Initialize the function estimator object.

        Args:
            enopt_config:    The configuration of the optimizer.
            estimator_index: The index of the estimator to use.
        """



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING

import numpy as np

from ropt.enums import ResultAxis

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig
    from ropt.transforms import OptModelTransforms


@dataclass(slots=True)
class ConstraintInfo(ResultField):
    """Stores information about constraint differences and violations.

     The `ConstraintInfo` class stores the differences between variable or
     constraint values and their respective bounds. It also calculates and
     stores constraint violations. This information is useful for assessing
     how well the optimization process is satisfying the imposed constraints.

     The class stores the following information:

     * **Bound Differences:** The differences between the variable values and
       their lower and upper bounds.
        - `bound_lower`: The difference between the variable values and their
          lower bounds. A negative value indicates that the variable is below its
          lower bound.
        - `bound_upper`: The difference between the variable values and their
          upper bounds. A positive value indicates that the variable is above its
          upper bound.

    * **Linear Constraint Differences:** The differences between the linear
       constraint values and their lower and upper bounds.
        - `linear_lower`: The difference between the linear constraint values
          and their lower bounds. A negative value indicates that the constraint
          is below its lower bound.
        - `linear_upper`: The difference between the linear constraint values and
          their upper bounds. A positive value indicates that the constraint is
          above its upper bound.

     * **Nonlinear Constraint Differences:** The differences between the
       nonlinear constraint values and their lower and upper bounds.
        - `nonlinear_lower`: The difference between the nonlinear constraint
          values and their lower bounds. A negative value indicates that the
          constraint is below its lower bound.
        - `nonlinear_upper`: The difference between the nonlinear constraint
          values and their upper bounds. A positive value indicates that the
          constraint is above its upper bound.

     * **Constraint Violations:** The magnitude of the constraint violations.
        - `bound_violation`: The magnitude of the violation of the variable
          bounds.
        - `linear_violation`: The magnitude of the violation of the linear
           constraints.
        - `nonlinear_violation`: The magnitude of the violation of the nonlinear
          constraints.

    Attributes:
         bound_lower:         Difference between variables and their lower bounds.
         bound_upper:         Difference between variables and their upper bounds.
         linear_lower:        Difference between linear constraints and their lower
                              bounds.
         linear_upper:        Difference between linear constraints and their upper
                              bounds.
         nonlinear_lower:     Difference between nonlinear constraints and their
                              lower bounds.
         nonlinear_upper:     Difference between nonlinear constraints and their
                              upper bounds.
         bound_violation:     Magnitude of the violation of the variable bounds.
         linear_violation:    Magnitude of the violation of the linear constraints.
         nonlinear_violation: Magnitude of the violation of the nonlinear constraints.
    """

    bound_lower: NDArray[np.float64] | None = field(
        default=None, metadata={"__axes__": (ResultAxis.VARIABLE,)}
    )
    bound_upper: NDArray[np.float64] | None = field(
        default=None, metadata={"__axes__": (ResultAxis.VARIABLE,)}
    )
    linear_lower: NDArray[np.float64] | None = field(
        default=None, metadata={"__axes__": (ResultAxis.LINEAR_CONSTRAINT,)}
    )
    linear_upper: NDArray[np.float64] | None = field(
        default=None, metadata={"__axes__": (ResultAxis.LINEAR_CONSTRAINT,)}
    )
    nonlinear_lower: NDArray[np.float64] | None = field(
        default=None, metadata={"__axes__": (ResultAxis.NONLINEAR_CONSTRAINT,)}
    )
    nonlinear_upper: NDArray[np.float64] | None = field(
        default=None, metadata={"__axes__": (ResultAxis.NONLINEAR_CONSTRAINT,)}
    )
    bound_violation: NDArray[np.float64] | None = field(
        default=None, metadata={"__axes__": (ResultAxis.VARIABLE,)}
    )
    linear_violation: NDArray[np.float64] | None = field(
        default=None, metadata={"__axes__": (ResultAxis.LINEAR_CONSTRAINT,)}
    )
    nonlinear_violation: NDArray[np.float64] | None = field(
        default=None, metadata={"__axes__": (ResultAxis.NONLINEAR_CONSTRAINT,)}
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.bound_lower = _immutable_copy(self.bound_lower)
        self.bound_upper = _immutable_copy(self.bound_upper)
        self.linear_lower = _immutable_copy(self.linear_lower)
        self.linear_upper = _immutable_copy(self.linear_upper)
        self.nonlinear_lower = _immutable_copy(self.nonlinear_lower)
        self.nonlinear_upper = _immutable_copy(self.nonlinear_upper)
        if self.bound_lower is not None and self.bound_upper is not None:
            self.bound_violation = _immutable_copy(
                np.maximum(
                    np.where(self.bound_lower < 0.0, -self.bound_lower, 0.0),
                    np.where(self.bound_upper > 0.0, self.bound_upper, 0.0),
                )
            )
        if self.linear_lower is not None and self.linear_upper is not None:
            self.linear_violation = _immutable_copy(
                np.maximum(
                    np.where(self.linear_lower < 0.0, -self.linear_lower, 0.0),
                    np.where(self.linear_upper > 0.0, self.linear_upper, 0.0),
                )
            )
        if self.nonlinear_lower is not None and self.nonlinear_upper is not None:
            self.nonlinear_violation = _immutable_copy(
                np.maximum(
                    np.where(self.nonlinear_lower < 0.0, -self.nonlinear_lower, 0.0),
                    np.where(self.nonlinear_upper > 0.0, self.nonlinear_upper, 0.0),
                )
            )

    @classmethod
    def create(
        cls,
        config: EnOptConfig,
        variables: NDArray[np.float64],
        constraints: NDArray[np.float64] | None,
    ) -> ConstraintInfo | None:
        """Add constraint difference information.

        This stores the following constraint differences:
        1. The difference between variables and their lower and upper bounds.
        2. The difference between the linear constraints with their
           right-hand-side upper and lower bounds, calculated for given variables.
        3. The differences between non-linear constraint values and their
           right-hand-side upper and lower bounds.

        Args:
            config:      The ensemble optimizer configuration object.
            variables:   The variables to check.
            constraints: The constraints to check (optional).

        Returns:
            A newly created ConstraintInfo object or None.
        """
        diffs: dict[str, NDArray[np.float64] | None] = {}

        if np.all(np.isfinite(config.variables.lower_bounds)) or np.all(
            np.isfinite(config.variables.upper_bounds)
        ):
            diffs["bound_lower"] = variables - config.variables.lower_bounds
            diffs["bound_upper"] = variables - config.variables.upper_bounds

        if config.linear_constraints is not None:
            values = np.matmul(config.linear_constraints.coefficients, variables)
            diffs["linear_lower"] = values - config.linear_constraints.lower_bounds
            diffs["linear_upper"] = values - config.linear_constraints.upper_bounds

        if constraints is not None:
            assert config.nonlinear_constraints is not None
            diffs["nonlinear_lower"] = (
                constraints - config.nonlinear_constraints.lower_bounds
            )
            diffs["nonlinear_upper"] = (
                constraints - config.nonlinear_constraints.upper_bounds
            )

        if diffs:
            return ConstraintInfo(**diffs)

        return None

    def transform_from_optimizer(
        self, transforms: OptModelTransforms
    ) -> ConstraintInfo:
        if transforms.variables is None and transforms.nonlinear_constraints is None:
            return self

        diffs: dict[str, NDArray[np.float64] | None] = asdict(self)

        if transforms.variables is not None and self.bound_lower is not None:
            assert self.bound_upper is not None
            diffs["bound_lower"], diffs["bound_upper"] = (
                transforms.variables.bound_constraint_diffs_from_optimizer(
                    self.bound_lower, self.bound_upper
                )
            )

        if transforms.variables is not None and self.linear_lower is not None:
            assert self.linear_upper is not None
            diffs["linear_lower"], diffs["linear_upper"] = (
                transforms.variables.linear_constraints_diffs_from_optimizer(
                    self.linear_lower, self.linear_upper
                )
            )

        if transforms.nonlinear_constraints is not None:
            assert self.nonlinear_lower is not None
            assert self.nonlinear_upper is not None
            diffs["nonlinear_lower"], diffs["nonlinear_upper"] = (
                transforms.nonlinear_constraints.nonlinear_constraint_diffs_from_optimizer(
                    self.nonlinear_lower, self.nonlinear_upper
                )
            )

        return ConstraintInfo(**diffs)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeVar

from ._results import Results

if TYPE_CHECKING:
    from ropt.transforms import OptModelTransforms

    from ._constraint_info import ConstraintInfo
    from ._function_evaluations import FunctionEvaluations
    from ._functions import Functions
    from ._realizations import Realizations


TypeResults = TypeVar("TypeResults", bound="Results")


@dataclass(slots=True)
class FunctionResults(Results):
    """Stores results related to function evaluations.

    The `FunctionResults` class extends the base
    [`Results`][ropt.results.Results] class to store data specific to function
    evaluations. This includes:

    * **Evaluations:** The results of the function evaluations, including the
      variable values, objective values, and constraint values for each
      realization. See
      [`FunctionEvaluations`][ropt.results.FunctionEvaluations].
    * **Realizations:** Information about the realizations, such as weights for
      objectives and constraints, and whether each realization was successful.
      See [`Realizations`][ropt.results.Realizations].
    * **Functions:** The calculated objective and constraint function values,
      typically aggregated across realizations. See
      [`Functions`][ropt.results.Functions].
    * **Constraint Info:** Details about constraint differences and violations.
      See [`ConstraintInfo`][ropt.results.ConstraintInfo].

    Attributes:
        evaluations:     Results of the function evaluations.
        realizations:    The calculated parameters of the realizations.
        functions:       The calculated functions.
        constraint_info: Information on constraint differences and violations.
    """

    evaluations: FunctionEvaluations
    realizations: Realizations
    functions: Functions | None
    constraint_info: ConstraintInfo | None = None

    def transform_from_optimizer(
        self, transforms: OptModelTransforms
    ) -> FunctionResults:
        """Apply transformations from optimizer space.

        Args:
            transforms: The transforms to apply.

        Returns:
            The transformed results.
        """
        return FunctionResults(
            batch_id=self.batch_id,
            metadata=self.metadata,
            evaluations=self.evaluations.transform_from_optimizer(transforms),
            realizations=self.realizations,
            functions=(
                None
                if self.functions is None
                else self.functions.transform_from_optimizer(transforms)
            ),
            constraint_info=(
                None
                if self.constraint_info is None
                else self.constraint_info.transform_from_optimizer(transforms)
            ),
        )



"""This module defines the optimization plan object."""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Self

from ropt.enums import EventType
from ropt.plugins import PluginManager

if TYPE_CHECKING:
    from ropt.evaluator import Evaluator
    from ropt.plan import Event


class OptimizerContext:
    """Manages shared state and resources for an optimization plan.

    The OptimizerContext acts as a central hub for managing shared resources and
    state across all steps within an optimization plan. This ensures that
    different parts of the plan can access and interact with the same
    information and tools.

    This context object is responsible for:

    - Providing a callable `Evaluator` for evaluating functions, which is
      essential for optimization algorithms to assess the quality of solutions.
      The evaluator is used to calculate the objective function's value and
      potentially constraint values for given variables.
    - Managing a `PluginManager` to retrieve and utilize plugins, allowing for
      extensibility and customization of the optimization workflow. Plugins are
      modular pieces of code that extend the functionality of the optimization
      framework, such as new `PlanStep` or `ResultHandler` implementations.
    - Handling event callbacks that are triggered in response to specific events
      during the plan's execution. These callbacks are executed after the plan
      has processed the event, allowing for actions to be taken in response to
      changes or milestones. This allows you to monitor the optimization, react
      to changes, or perform custom actions at specific points in the plan's
      execution.

    Args:
        evaluator:      A callable for evaluating functions in the plan.
        plugin_manager: A plugin manager; a default is created if not provided.
    """

    def __init__(
        self,
        evaluator: Evaluator,
        plugin_manager: PluginManager | None = None,
    ) -> None:
        """Initializes the optimization context.

        Sets up the shared state and resources required for an optimization
        plan. This includes a function evaluator and a plugin manager.

        Args:
            evaluator:      A callable for evaluating functions in the plan.
            plugin_manager: A plugin manager; a default is created if not provided.
        """
        self.evaluator = evaluator
        self.plugin_manager = (
            PluginManager() if plugin_manager is None else plugin_manager
        )
        self._subscribers: dict[EventType, list[Callable[[Event], None]]] = {
            event: [] for event in EventType
        }

    def add_observer(
        self,
        event: EventType,
        callback: Callable[[Event], None],
    ) -> Self:
        """Adds an observer function for a specific event type.

        Observer functions are called when an event of the specified type
        occurs during optimization. The provided callback function will receive
        an [`Event`][ropt.plan.Event] object containing information about the
        event.

        Args:
            event:    The type of event to observe.
            callback: The function to call when the event occurs.

        Returns:
            The OptimizerContext instance, allowing for method chaining.
        """
        self._subscribers[event].append(callback)
        return self

    def call_observers(self, event: Event) -> None:
        """Calls all observers for a specific event.

        This method triggers all observer functions registered for the given
        event type.

        Args:
            event: The event to emit to the observers.
        """
        for callback in self._subscribers[event.event_type]:
            callback(event)



"""Configuration class for non-linear constraints."""

from __future__ import annotations

from typing import Self

import numpy as np
from pydantic import ConfigDict, ValidationInfo, model_validator

from ropt.config.utils import ImmutableBaseModel, broadcast_arrays, immutable_array
from ropt.config.validated_types import (  # noqa: TC001
    Array1D,
    Array1DInt,
)


class NonlinearConstraintsConfig(ImmutableBaseModel):
    r"""Configuration class for non-linear constraints.

    This class, `NonlinearConstraintsConfig`, defines non-linear constraints used
    in an [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.

    Non-linear constraints are defined by comparing a constraint function to a
    right-hand-side value, allowing for equality or inequality constraints. The
    `lower_bounds` and `upper_bounds` fields, which are `numpy` arrays, specify the
    bounds on these right-hand-side values. The length of these arrays determines
    the number of constraint functions.

    Less-than and greater-than inequality constraints can be specified by
    setting the lower bounds to $-\infty$, or the upper bounds to $+\infty$,
    respectively. Equality constraints are specified by setting the lower bounds
    equal to the upper bounds.

    Non-linear constraints can be processed by realization filters and function
    estimators. The `realization_filters` and `function_estimators` fields contain
    indices that refer to the corresponding objects configured in the parent
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.

    Attributes:
        lower_bounds:        Lower bounds for the right-hand-side values.
        upper_bounds:        Upper bounds for the right-hand-side values.
        realization_filters: Optional indices of realization filters.
        function_estimators: Optional indices of function estimators.
    """

    lower_bounds: Array1D
    upper_bounds: Array1D
    realization_filters: Array1DInt | None = None
    function_estimators: Array1DInt | None = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _broadcast_and_check(self, info: ValidationInfo) -> Self:
        lower_bounds, upper_bounds = broadcast_arrays(
            self.lower_bounds, self.upper_bounds
        )

        if info.context is not None and info.context.nonlinear_constraints is not None:
            lower_bounds, upper_bounds = (
                info.context.nonlinear_constraints.bounds_to_optimizer(
                    lower_bounds, upper_bounds
                )
            )

        if np.any(lower_bounds > upper_bounds):
            msg = "The non-linear constraint lower bounds are larger than the upper bounds."
            raise ValueError(msg)

        self._mutable()
        self.lower_bounds = immutable_array(lower_bounds)
        self.upper_bounds = immutable_array(upper_bounds)
        self._immutable()

        return self



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ropt.enums import ResultAxis

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray


@dataclass(slots=True)
class Realizations(ResultField):
    """Stores information about the realizations.

    The `Realizations` class stores data related to the individual realizations
    used in the optimization process. This includes:

    * **Failed Realizations:** A boolean array indicating whether each
      realization's evaluation was successful. `True` indicates a failed
      realization, while `False` indicates a successful one.
    * **Objective Weights:** A two-dimensional array of weights used for each
      objective in each realization. The first axis corresponds to the
      objectives, and the second axis corresponds to the realizations. These
      weights may change during optimization, depending on the type of objective
      calculation.
    * **Constraint Weights:** A two-dimensional array of weights used for each
      constraint in each realization. The first axis corresponds to the
      constraints, and the second axis corresponds to the realizations. These
      weights may change during optimization, depending on the type of
      constraint calculation.

    Attributes:
        failed_realizations: Boolean array indicating failed realizations.
        objective_weights:   Weights for each objective in each realization.
        constraint_weights:  Weights for each constraint in each realization.
    """

    failed_realizations: NDArray[np.bool_] = field(
        metadata={
            "__axes__": (ResultAxis.REALIZATION,),
        },
    )
    objective_weights: NDArray[np.float64] | None = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxis.OBJECTIVE,
                ResultAxis.REALIZATION,
            ),
        },
    )
    constraint_weights: NDArray[np.float64] | None = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxis.NONLINEAR_CONSTRAINT,
                ResultAxis.REALIZATION,
            ),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.failed_realizations = _immutable_copy(self.failed_realizations)
        self.objective_weights = _immutable_copy(self.objective_weights)
        self.constraint_weights = _immutable_copy(self.constraint_weights)



from typing import Any

import numpy as np
import pytest

from ropt.enums import ResultAxis
from ropt.results import (
    FunctionEvaluations,
    FunctionResults,
    Functions,
    GradientEvaluations,
    GradientResults,
    Gradients,
    Realizations,
)

pandas = pytest.importorskip("pandas")
from ropt.results._pandas import _to_series


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> dict[str, Any]:
    return {
        "variables": {
            "initial_values": [0.0, 0.0],
        },
        "objectives": {
            "weights": [0.75, 0.25],
        },
        "realizations": {
            "weights": [1.0] * 3,
        },
        "gradient": {
            "number_of_perturbations": 5,
        },
    }


@pytest.fixture(name="function_result")
def function_result_fixture() -> FunctionResults:
    evaluations = FunctionEvaluations.create(
        variables=np.array([1.0, 2.0]),
        objectives=np.arange(6, dtype=np.float64).reshape((3, 2)),
    )
    realizations = Realizations(
        objective_weights=np.arange(6, dtype=np.float64).reshape((2, 3)),
        failed_realizations=np.zeros(3, dtype=np.bool_),
    )
    functions = Functions.create(
        weighted_objective=np.array(1.0), objectives=np.array([1.0, 2.0])
    )
    return FunctionResults(
        batch_id=1,
        metadata={},
        evaluations=evaluations,
        realizations=realizations,
        functions=functions,
    )


@pytest.fixture(name="gradient_result")
def gradient_result_fixture() -> GradientResults:
    evaluations = GradientEvaluations(
        variables=np.array([1.0, 2.0]),
        perturbed_variables=np.arange(30, dtype=np.float64).reshape((3, 5, 2)),
        perturbed_objectives=np.arange(30, dtype=np.float64).reshape((3, 5, 2)),
        evaluation_info={"foo": np.arange(15, dtype=np.float64).reshape((3, 5))},
    )
    gradients = Gradients(
        weighted_objective=np.array([1.0, 2.0]),
        objectives=np.arange(4, dtype=np.float64).reshape((2, 2)),
    )
    return GradientResults(
        batch_id=1,
        metadata={},
        evaluations=evaluations,
        realizations=Realizations(
            failed_realizations=np.zeros(36, dtype=np.bool_),
        ),
        gradients=gradients,
    )


def test__to_series(gradient_result: GradientResults) -> None:
    names: dict[str, Any] = {
        ResultAxis.VARIABLE: ("v1", "v2"),
        ResultAxis.REALIZATION: ("ra", "rb", "rc"),
    }
    series = _to_series(gradient_result.evaluations, "perturbed_variables", None, names)
    assert series is not None
    assert len(series) == gradient_result.evaluations.perturbed_variables.size
    assert series.index.names == [
        "realization",
        "perturbation",
        "variable",
    ]
    for v_idx, var in enumerate(names[ResultAxis.VARIABLE]):
        for r_idx, real in enumerate(names[ResultAxis.REALIZATION]):
            for pert in range(gradient_result.evaluations.perturbed_variables.shape[1]):
                assert (
                    series.loc[(real, pert, var)]
                    == gradient_result.evaluations.perturbed_variables[
                        r_idx, pert, v_idx
                    ]
                )


def test__to_series_evaluation_info(gradient_result: GradientResults) -> None:
    names: dict[str, Any] = {
        ResultAxis.VARIABLE: ("v1", "v2"),
        ResultAxis.REALIZATION: ("ra", "rb", "rc"),
    }
    series = _to_series(gradient_result.evaluations, "evaluation_info", "foo", names)
    assert series is not None
    info = np.array(gradient_result.evaluations.evaluation_info["foo"])
    assert len(series) == info.size
    assert series.index.names == [
        "realization",
        "perturbation",
    ]
    for r_idx, real in enumerate(names[ResultAxis.REALIZATION]):
        for pert in range(gradient_result.evaluations.perturbed_variables.shape[1]):
            assert series.loc[(real, pert)] == info[r_idx, pert]


def test_to_dataframe_function(function_result: FunctionResults) -> None:
    names: dict[str, Any] = {
        ResultAxis.VARIABLE: ("v1", "v2"),
        ResultAxis.REALIZATION: ("ra", "rb", "rc"),
        ResultAxis.OBJECTIVE: ("fa", "fb"),
    }
    frame = function_result.to_dataframe(
        "functions",
        [
            "weighted_objective",
            "objectives",
        ],
        names=names,
    )
    assert len(frame) == 2
    assert frame.index.names == ["batch_id", "objective"]
    assert frame.index[0] == (1, "fa")
    assert frame.index[1] == (1, "fb")


def test_to_dataframe_gradient(gradient_result: GradientResults) -> None:
    names: dict[str, Any] = {
        ResultAxis.VARIABLE: ("v1", "v2"),
        ResultAxis.REALIZATION: ("ra", "rb", "rc"),
        ResultAxis.OBJECTIVE: ("fa", "fb"),
    }
    frame = gradient_result.to_dataframe(
        "evaluations",
        [
            "variables",
            "perturbed_variables",
            "perturbed_objectives",
            "evaluation_info.foo",
        ],
        names=names,
    )
    assert len(frame) == gradient_result.evaluations.perturbed_variables.size * 2
    assert frame.index.names == [
        "batch_id",
        "variable",
        "realization",
        "perturbation",
        "objective",
    ]
    idx = 0
    for var in names[ResultAxis.VARIABLE]:
        for real in names[ResultAxis.REALIZATION]:
            for pert in range(gradient_result.evaluations.perturbed_variables.shape[1]):
                for fnc in names[ResultAxis.OBJECTIVE]:
                    assert frame.index[idx] == (1, var, real, pert, fnc)
                    idx += 1


def test_to_dataframe_unstack1(gradient_result: GradientResults) -> None:
    names: dict[str, Any] = {
        ResultAxis.VARIABLE: ("x", "y"),
        ResultAxis.REALIZATION: (2, 3, 1),
    }
    frame = gradient_result.to_dataframe(
        "evaluations",
        select=["perturbed_variables"],
        unstack=[ResultAxis.REALIZATION, ResultAxis.VARIABLE],
        names=names,
    )
    assert frame.index.names == ["batch_id", "perturbation"]
    assert list(frame.columns.values) == [
        ("perturbed_variables", 2, "x"),
        ("perturbed_variables", 2, "y"),
        ("perturbed_variables", 3, "x"),
        ("perturbed_variables", 3, "y"),
        ("perturbed_variables", 1, "x"),
        ("perturbed_variables", 1, "y"),
    ]


def test_to_dataframe_unstack2(gradient_result: GradientResults) -> None:
    names: dict[str, Any] = {
        ResultAxis.VARIABLE: ("x", "y"),
        ResultAxis.OBJECTIVE: ("f1", "f2"),
    }
    assert gradient_result.gradients is not None
    frame = gradient_result.to_dataframe(
        "gradients",
        select=["objectives", "weighted_objective"],
        unstack=[ResultAxis.OBJECTIVE, ResultAxis.VARIABLE],
        names=names,
    )
    assert list(frame.columns.values) == [
        ("objectives", "f1", "x"),
        ("objectives", "f1", "y"),
        ("objectives", "f2", "x"),
        ("objectives", "f2", "y"),
        ("weighted_objective", "x"),
        ("weighted_objective", "y"),
    ]


def test_to_dataframe_unstack_only_variable(gradient_result: GradientResults) -> None:
    names: dict[str, Any] = {ResultAxis.VARIABLE: ("x", "y")}
    frame = gradient_result.to_dataframe(
        "evaluations",
        select=["perturbed_objectives", "perturbed_variables"],
        unstack=[ResultAxis.VARIABLE],
        names=names,
    )
    assert frame.index.names == [
        "batch_id",
        "realization",
        "perturbation",
        "objective",
    ]
    assert list(frame.columns.values) == [
        "perturbed_objectives",
        ("perturbed_variables", "x"),
        ("perturbed_variables", "y"),
    ]


def test_to_dataframe_join(function_result: FunctionResults) -> None:
    frame1 = function_result.to_dataframe("evaluations", ["variables", "objectives"])
    frame2 = function_result.to_dataframe(
        "functions", ["weighted_objective", "objectives"]
    )
    frame1.columns = pandas.Index(
        "_".join(column) if isinstance(column, tuple) else column
        for column in frame1.columns.to_numpy()
    )
    frame2.columns = pandas.Index(
        "_".join(column) if isinstance(column, tuple) else column
        for column in frame2.columns.to_numpy()
    )
    frame = frame1.join(frame2, how="inner", lsuffix="_eval", rsuffix="_func")
    assert not frame.empty
    assert len(frame.columns) == len(frame1.columns) + len(frame2.columns)



"""Ensemble optimizer class."""

from __future__ import annotations

import os
import sys
from contextlib import contextmanager
from typing import TYPE_CHECKING, Generator, Protocol

import numpy as np

from ropt.enums import OptimizerExitCode
from ropt.exceptions import OptimizationAborted
from ropt.results import FunctionResults, GradientResults

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig
    from ropt.ensemble_evaluator import EnsembleEvaluator
    from ropt.plugins import PluginManager
    from ropt.plugins.optimizer.base import Optimizer
    from ropt.results import Functions, Gradients, Results


class SignalEvaluationCallback(Protocol):
    """Protocol for a callback to signal the start and end of an evaluation.

    This callback is invoked before and after each evaluation, allowing for
    custom handling or tracking of evaluation events.
    """

    def __call__(
        self,
        results: tuple[Results, ...] | None = None,
    ) -> None:
        """Callback protocol for signaling the start and end of evaluations.

        This callback is invoked by the ensemble optimizer before and after
        each evaluation. Before the evaluation starts, the callback is called
        with `results` set to `None`. After the evaluation completes, the
        callback is called again, this time with `results` containing the
        output of the evaluation.

        Args:
            results: The results produced by the evaluation, or `None` if the
                     evaluation has not yet started.
        """


class NestedOptimizerCallback(Protocol):
    """Protocol for functions that start a nested optimization."""

    def __call__(
        self, variables: NDArray[np.float64], /
    ) -> tuple[FunctionResults | None, bool]:
        """Callback protocol for executing a nested optimization.

        This function is invoked by the ensemble optimizer during each function
        evaluation to initiate a nested optimization process. It receives the
        current variables as input and is expected to perform a nested
        optimization using these variables as a starting point. The function
        should return a tuple containing the result of the nested optimization
        and a boolean indicating whether the nested optimization was aborted by
        the user. The result of the nested optimization should be a
        [`FunctionResults`][ropt.results.FunctionResults] object, or `None` if
        no result is available.

        Args:
            variables: The variables to use as the starting point.

        Returns:
            The result and a boolean indicating if the user aborted.
        """


class EnsembleOptimizer:
    """Optimizer for ensemble-based optimizations.

    The [`EnsembleOptimizer`][ropt.optimization.EnsembleOptimizer] class
    provides the core functionality for running ensemble-based optimizations.
    Direct use of this class is generally discouraged. Instead, the
    [`Plan`][ropt.plan.Plan] or [`BasicOptimizer`][ropt.plan.BasicOptimizer]
    classes are recommended for greater flexibility and ease of use.
    """

    def __init__(
        self,
        enopt_config: EnOptConfig,
        ensemble_evaluator: EnsembleEvaluator,
        plugin_manager: PluginManager,
        signal_evaluation: SignalEvaluationCallback | None = None,
        nested_optimizer: NestedOptimizerCallback | None = None,
    ) -> None:
        """Initialize the EnsembleOptimizer.

        This class orchestrates ensemble-based optimizations. It requires an
        optimization configuration, an evaluator, and a plugin manager to
        function.

        The `EnsembleOptimizer` needs the following to define a single
        optimization run:

        1.  An [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object: This
            contains all configuration settings for the optimization.
        2.  An [`EnsembleEvaluator`][ropt.ensemble_evaluator.EnsembleEvaluator]
            object: This object is responsible for evaluating functions.
        3.  A [`PluginManager`][ropt.plugins.PluginManager] object: This object
            provides access to optimizer plugins.

        Additionally, two optional callbacks can be provided to extend the
        functionality:

        1.  A
            [`SignalEvaluationCallback`][ropt.optimization.SignalEvaluationCallback]:
            This callback is invoked before and after each function evaluation.
        2.  A
            [`NestedOptimizerCallback`][ropt.optimization.NestedOptimizerCallback]:
            This callback is invoked at each function evaluation to run a nested
            optimization.

        Args:
            enopt_config:       The ensemble optimization configuration.
            ensemble_evaluator: The evaluator for function evaluations.
            plugin_manager:     The plugin manager.
            signal_evaluation:  Optional callback to signal evaluations.
            nested_optimizer:   Optional callback for nested optimizations.
        """
        self._enopt_config = enopt_config
        self._function_evaluator = ensemble_evaluator
        self._plugin_manager = plugin_manager
        self._signal_evaluation = signal_evaluation
        self._nested_optimizer = nested_optimizer

        # This stores the values of the fixed variable
        self._fixed_variables: NDArray[np.float64]

        # For implementing max_functions:
        self._completed_functions = 0

        # Whether NaN values are allowed:
        self._allow_nan = False
        self._optimizer: Optimizer = self._plugin_manager.get_plugin(
            "optimizer", method=self._enopt_config.optimizer.method
        ).create(self._enopt_config, self._optimizer_callback)
        self._allow_nan = self._optimizer.allow_nan

        # Optional redirection of standard output:
        self._redirector = _Redirector(self._enopt_config)

    @property
    def is_parallel(self) -> bool:
        """Determine if the optimization supports parallel evaluations.

        The underlying optimization algorithm may request function evaluations
        via a callback. Parallel optimization, in this context, means that the
        algorithm may request multiple function evaluations in a single
        callback.

        Returns:
            `True` if the optimization supports parallel evaluations, `False`
            otherwise.
        """
        return self._optimizer.is_parallel

    def start(self, variables: NDArray[np.float64]) -> OptimizerExitCode:
        """Start the optimization process.

        This method initiates the optimization process using the provided
        initial variables. The optimization will continue until a stopping
        criterion is met or an error occurs.

        Args:
            variables: The initial variables for the optimization.

        Returns:
            An [`OptimizerExitCode`][ropt.enums.OptimizerExitCode] indicating
            the reason for termination.
        """
        self._fixed_variables = variables.copy()
        exit_code = OptimizerExitCode.OPTIMIZER_STEP_FINISHED
        try:
            with self._redirector.start():
                self._optimizer.start(variables)
        except OptimizationAborted as exc:
            exit_code = exc.exit_code
        return exit_code

    def _optimizer_callback(
        self,
        variables: NDArray[np.float64],
        *,
        return_functions: bool,
        return_gradients: bool,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        assert return_functions or return_gradients

        self._check_stopping_criteria()

        variables = self._get_completed_variables(variables)

        # Run any nested steps, when this improves the objective, this may
        # change the fixed variables and the current optimal result:
        if self._nested_optimizer is not None:
            # Nested optimization does not support parallel evaluation:
            if variables.ndim > 1:
                if variables.shape[0] > 1:
                    msg = "Nested optimization does not support parallel evaluation."
                    raise RuntimeError(msg)
                variables = variables[0, ...]
            nested_results, aborted = self._nested_optimizer(variables)
            if aborted:
                raise OptimizationAborted(exit_code=OptimizerExitCode.USER_ABORT)
            if nested_results is None:
                raise OptimizationAborted(
                    exit_code=OptimizerExitCode.NESTED_OPTIMIZER_FAILED
                )
            variables = nested_results.evaluations.variables
            self._fixed_variables = variables.copy()

        results = self._run_evaluations(
            variables,
            compute_functions=return_functions,
            compute_gradients=return_gradients,
        )

        functions = np.array([])
        if return_functions:
            # Functions might be parallelized hence we need potentially to
            # process a list of function results:
            functions_list = [
                self._functions_from_results(item.functions)
                for item in results
                if isinstance(item, FunctionResults)
            ]
            self._completed_functions += len(functions_list)
            functions = (
                np.vstack(functions_list) if variables.ndim > 1 else functions_list[0]
            )

        gradients = np.array([])
        if return_gradients:
            # Gradients cannot be parallelized, there is at most one gradient:
            gradients = self._gradients_from_results(
                next(
                    item.gradients
                    for item in results
                    if isinstance(item, GradientResults)
                ),
                self._enopt_config.variables.mask,
            )

        return functions, gradients

    def _get_completed_variables(
        self, variables: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        mask = self._enopt_config.variables.mask
        if mask is not None:
            if variables.ndim > 1:
                tmp_variables = np.repeat(
                    self._fixed_variables[np.newaxis, :], variables.shape[0], axis=0
                )
                tmp_variables[:, mask] = variables
            else:
                tmp_variables = self._fixed_variables.copy()
                tmp_variables[mask] = variables
            return tmp_variables
        return variables.copy()

    def _check_stopping_criteria(self) -> None:
        max_functions = self._enopt_config.optimizer.max_functions
        if max_functions is not None and self._completed_functions >= max_functions:
            raise OptimizationAborted(exit_code=OptimizerExitCode.MAX_FUNCTIONS_REACHED)

    def _run_evaluations(
        self,
        variables: NDArray[np.float64],
        *,
        compute_functions: bool = False,
        compute_gradients: bool = False,
    ) -> tuple[Results, ...]:
        with self._redirector.suspend():
            assert compute_functions or compute_gradients
            if self._signal_evaluation:
                self._signal_evaluation()
            results = self._function_evaluator.calculate(
                variables,
                compute_functions=compute_functions,
                compute_gradients=compute_gradients,
            )

            # If the configuration allows for zero successful realizations, there
            # will always be results. However, they may all be equal to `np.nan`. If
            # the optimizer does not set the allow_nan flag, it cannot handle such a
            # case, and we need to check for it:
            assert self._enopt_config.realizations.realization_min_success is not None
            check_failures = (
                self._enopt_config.realizations.realization_min_success < 1
                and not self._allow_nan
            )
            exit_code: OptimizerExitCode | None = None
            for result in results:
                assert isinstance(result, FunctionResults | GradientResults)
                if (
                    (isinstance(result, FunctionResults) and result.functions is None)
                    or (
                        isinstance(result, GradientResults) and result.gradients is None
                    )
                    or (
                        check_failures
                        and np.all(result.realizations.failed_realizations)
                    )
                ):
                    exit_code = OptimizerExitCode.TOO_FEW_REALIZATIONS
                    break

            if self._signal_evaluation:
                self._signal_evaluation(results)

            if exit_code is not None:
                raise OptimizationAborted(exit_code=exit_code)

        return results

    @staticmethod
    def _functions_from_results(functions: Functions | None) -> NDArray[np.float64]:
        assert functions is not None
        return (
            np.array(functions.weighted_objective, ndmin=1)
            if functions.constraints is None
            else np.append(functions.weighted_objective, functions.constraints)
        )

    @staticmethod
    def _gradients_from_results(
        gradients: Gradients | None, mask: NDArray[np.bool_] | None
    ) -> NDArray[np.float64]:
        assert gradients is not None
        weighted_objective_gradient = (
            gradients.weighted_objective.copy()
            if mask is None
            else gradients.weighted_objective[mask]
        )
        constraint_gradients = (
            None
            if gradients.constraints is None
            else (
                gradients.constraints.copy()
                if mask is None
                else gradients.constraints[:, mask]
            )
        )
        return (
            np.expand_dims(weighted_objective_gradient, axis=0)
            if constraint_gradients is None
            else np.vstack((weighted_objective_gradient, constraint_gradients))
        )


class _Redirector:
    def __init__(self, config: EnOptConfig) -> None:
        output_dir = config.optimizer.output_dir
        stdout = config.optimizer.stdout
        stderr = config.optimizer.stderr

        if stdout is not None:
            self._redirect = True
            if stderr is None:
                stderr = stdout
            if not stdout.is_absolute() and output_dir is not None:
                stdout = output_dir / stdout
            if not stderr.is_absolute() and output_dir is not None:
                stderr = output_dir / stderr
            sys.stdout.flush()
            sys.stderr.flush()
            self._old_stdout = os.dup(1)
            self._old_stderr = os.dup(2)
            self._new_stdout = os.open(stdout, os.O_WRONLY | os.O_CREAT)
            self._new_stderr = os.open(stderr, os.O_WRONLY | os.O_CREAT)
        else:
            self._redirect = False

    @contextmanager
    def start(self) -> Generator[None, None, None]:
        if self._redirect:
            try:
                sys.stdout.flush()
                sys.stderr.flush()
                os.dup2(self._new_stdout, 1)
                os.dup2(self._new_stderr, 2)
                yield
            finally:
                os.dup2(self._old_stdout, 1)
                os.dup2(self._old_stderr, 2)
                os.close(self._new_stdout)
                os.close(self._new_stderr)
        else:
            yield

    @contextmanager
    def suspend(self) -> Generator[None, None, None]:
        if self._redirect:
            try:
                os.fsync(self._new_stdout)
                os.fsync(self._new_stderr)
                os.dup2(self._old_stdout, 1)
                os.dup2(self._old_stderr, 2)
                yield
            finally:
                sys.stdout.flush()
                sys.stderr.flush()
                os.dup2(self._new_stdout, 1)
                os.dup2(self._new_stderr, 2)
        else:
            yield



"""This module implements the default evaluator step."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any

import numpy as np

from ropt.config.enopt import EnOptConfig
from ropt.ensemble_evaluator import EnsembleEvaluator
from ropt.enums import EventType, OptimizerExitCode
from ropt.exceptions import OptimizationAborted
from ropt.plan import Event
from ropt.plugins.plan.base import PlanStep
from ropt.results import FunctionResults

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

    from ropt.plan import Plan
    from ropt.transforms import OptModelTransforms


class DefaultEvaluatorStep(PlanStep):
    """The default evaluator step.

    This step performs a single ensemble evaluation, yielding one or more
    [`FunctionResults`][ropt.results.FunctionResults] objects. The evaluation
    can process multiple variable vectors, each of which is evaluated
    separately, producing an individual results object for each vector.

    Before executing the evaluator step, a
    [`START_EVALUATOR_STEP`][ropt.enums.EventType.START_EVALUATOR_STEP] event is
    emitted. After the evaluator step finishes, an
    [`FINISHED_EVALUATOR_STEP`][ropt.enums.EventType.FINISHED_EVALUATOR_STEP]
    event is emitted. Result handlers should respond to the latter event to
    process the generated results.
    """

    def __init__(
        self,
        plan: Plan,
    ) -> None:
        """Initialize a default evaluator step.

        Args:
            plan: The plan that runs this step.
        """
        super().__init__(plan)

    def run(  # type: ignore[override]
        self,
        config: dict[str, Any] | EnOptConfig,
        transforms: OptModelTransforms | None = None,
        variables: ArrayLike | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> OptimizerExitCode:
        """Run the evaluator step.

        The
        [`DefaultEvaluatorStep`][ropt.plugins.plan.evaluator.DefaultEvaluatorStep]
        requires an optimizer configuration; the `variables` parameter is
        optional. The configuration  object must be an
        [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object, or a dictionary
        that can be parsed into such an object. If no `variables` are provided,
        the initial values specified by the optimizer configuration are used. If
        `values` is given, it may be a single vector or a two-dimensional array.
        In the latter case, each row of the matrix is treated as a separate set
        of values to be evaluated.

        Args:
            config:     The optimizer configuration.
            transforms: Optional transforms object.
            variables: Variables to evaluate.
            metadata:   Optional metadata to add to events.
        """
        config = EnOptConfig.model_validate(config, context=transforms)

        self.emit_event(
            Event(
                event_type=EventType.START_EVALUATOR_STEP,
                config=config,
                source=self.id,
            )
        )

        if variables is None:
            variables = config.variables.initial_values
        else:
            variables = np.array(np.asarray(variables, dtype=np.float64), ndmin=1)
            if transforms is not None and transforms.variables is not None:
                variables = transforms.variables.to_optimizer(variables)

        ensemble_evaluator = EnsembleEvaluator(
            config,
            transforms,
            self.plan.optimizer_context.evaluator,
            self.plan.optimizer_context.plugin_manager,
        )

        exit_code = OptimizerExitCode.EVALUATION_STEP_FINISHED

        self.emit_event(
            Event(
                event_type=EventType.START_EVALUATION,
                config=config,
                source=self.id,
            )
        )
        try:
            results = ensemble_evaluator.calculate(
                variables, compute_functions=True, compute_gradients=False
            )
        except OptimizationAborted as exc:
            exit_code = exc.exit_code

        assert results
        assert isinstance(results[0], FunctionResults)
        if results[0].functions is None:
            exit_code = OptimizerExitCode.TOO_FEW_REALIZATIONS

        if metadata is not None:
            for item in results:
                item.metadata = deepcopy(metadata)

        data: dict[str, Any] = {"results": results}

        self.emit_event(
            Event(
                event_type=EventType.FINISHED_EVALUATION,
                config=config,
                source=self.id,
                data=data,
            )
        )

        if exit_code == OptimizerExitCode.USER_ABORT:
            self.plan.abort()

        self.emit_event(
            Event(
                event_type=EventType.FINISHED_EVALUATOR_STEP,
                config=config,
                source=self.id,
            )
        )

        return exit_code

    def emit_event(self, event: Event) -> None:
        """Emit an event.

        Args:
            event: The event to emit.
        """
        self.plan.emit_event(event)



"""The plugin manager."""

from __future__ import annotations

from functools import cache
from importlib.metadata import entry_points
from typing import TYPE_CHECKING, Any, Final, Generator, Literal

from ropt.exceptions import ConfigError

from .function_estimator.base import FunctionEstimatorPlugin
from .optimizer.base import OptimizerPlugin
from .plan.base import PlanHandlerPlugin, PlanStepPlugin
from .realization_filter.base import RealizationFilterPlugin
from .sampler.base import SamplerPlugin

if TYPE_CHECKING:
    from ropt.plugins.base import Plugin


_PLUGIN_TYPES: Final = {
    "function_estimator": FunctionEstimatorPlugin,
    "optimizer": OptimizerPlugin,
    "sampler": SamplerPlugin,
    "realization_filter": RealizationFilterPlugin,
    "plan_handler": PlanHandlerPlugin,
    "plan_step": PlanStepPlugin,
}

PluginType = Literal[
    "optimizer",
    "sampler",
    "realization_filter",
    "function_estimator",
    "plan_handler",
    "plan_step",
]
""" Plugin Types Supported by `ropt`"""


class PluginManager:
    """The plugin manager."""

    def __init__(self) -> None:
        """Initialize the plugin manager.

        The plugin manager object is initialized via the entry points mechanism
        (see
        [`importlib.metadata`](https://docs.python.org/3/library/importlib.metadata.html)).

        For instance, to install an additional optimizer plugin, implemented in
        an independent package, and assuming installation via a `pyproject.toml`
        file, add the following:

        ```toml
        [project.entry-points."ropt.plugins.optimizer"]
        my_optimizer = "my_optimizer_pkg.my_plugin:MyOptimizer"
        ```
        This will make the `MyOptimizer` class from the `my_optimizer_pkg`
        package available under the name `my_optimizer`. The `MyOptimizer` class
        will be used to create
        [`OptimizerPlugin`][ropt.plugins.optimizer.base.OptimizerPlugin] objects.
        """
        # Built-in plugins, listed for all possible plugin types:
        self._plugins: dict[PluginType, dict[str, type[Plugin]]] = {
            "optimizer": {},
            "sampler": {},
            "realization_filter": {},
            "function_estimator": {},
            "plan_handler": {},
            "plan_step": {},
        }

        for plugin_type in self._plugins:
            for name, plugin in _from_entry_points(plugin_type).items():
                self._add_plugin(plugin_type, name, plugin)

    def _add_plugin(
        self,
        plugin_type: PluginType,
        name: str,
        plugin: type[Plugin],
        *,
        prioritize: bool = False,
    ) -> None:
        name_lower = name.lower()
        if name_lower in self._plugins[plugin_type]:
            msg = f"Duplicate plugin name: {name_lower}"
            raise ConfigError(msg)
        if prioritize:
            plugins = self._plugins[plugin_type]
            self._plugins[plugin_type] = {name_lower: plugin}
            self._plugins[plugin_type].update(dict(plugins))
        else:
            self._plugins[plugin_type][name_lower] = plugin

    def _get_plugin(self, plugin_type: PluginType, method: str) -> Any | None:  # noqa: ANN401
        split_method = method.split("/", maxsplit=1)
        if len(split_method) > 1:
            plugin = self._plugins[plugin_type].get(split_method[0].lower())
            if plugin and plugin.is_supported(split_method[1]):
                return plugin
        else:
            if split_method[0] == "default":
                msg = "Cannot specify 'default' method without a plugin name"
                raise ConfigError(msg)
            for plugin in self._plugins[plugin_type].values():
                if plugin.allows_discovery() and plugin.is_supported(split_method[0]):
                    return plugin
        return None

    def get_plugin(self, plugin_type: PluginType, method: str) -> Any:  # noqa: ANN401
        """Retrieve a plugin by type and method name.

        If the method name is of the form "_plugin-name/method-name_", the method
        _method-name_ will be retrieved from the given plugin _plugin-name_.

        If the given method name does not contain a slash (/), the plugin
        manager will search through all plugins and return the first plugin that
        supports the requested method. Searching occurs in the order that plugins
        were added to the manager, which normally will be one of the plugins loaded
        via entry points, but plugins added dynamically can be prioritized.

        Args:
            plugin_type: The type of the plugin to retrieve.
            method:      The name of the method the plugin should provide.
        """
        if (plugin := self._get_plugin(plugin_type, method)) is not None:
            return plugin
        msg = f"Method not found: {method}"
        raise ConfigError(msg)

    def is_supported(self, plugin_type: PluginType, method: str) -> bool:
        """Check if a method is supported.

        If the given method name does not contain a slash (/), the plugin
        manager will search through all plugins and return `True` if a plugin is
        found that supports the requested method. If the method name is of the
        form "_plugin-name/method-name_", `True` will be returned if the method
        _method-name_ is supported by the given plugin _plugin-name_.

        Args:
            plugin_type: The type of the plugin to retrieve.
            method:      The name of the method the plugin should provide.
        """
        return self._get_plugin(plugin_type, method) is not None

    def plugins(
        self, plugin_type: PluginType
    ) -> Generator[tuple[str, type[Plugin]], None, None]:
        """Generate a sequence of all plugins of a specified type.

        Args:
            plugin_type: The type of plugins to return.

        Yields:
            A tuple of the plugin name and object.
        """
        for plugin_name, plugin in self._plugins[plugin_type].items():
            yield plugin_name, plugin


@cache  # Without the cache, repeated calls are very slow
def _from_entry_points(plugin_type: str) -> dict[str, type[Plugin]]:
    plugins: dict[str, type[Plugin]] = {}
    for entry_point in entry_points().select(group=f"ropt.plugins.{plugin_type}"):
        plugin = entry_point.load()
        plugins[entry_point.name] = plugin
        if not issubclass(plugins[entry_point.name], _PLUGIN_TYPES[plugin_type]):
            msg = (
                f"Incorrect type for {plugin_type} plugin `{entry_point.name}`"
                f": {type(plugins[entry_point.name])}"
            )
            raise TypeError(msg)
    return plugins



"""Differential evaluation optimization example.

This example uses the differential evolution method to solve a discrete
problem with a linear constraint.
"""

from typing import Any

import numpy as np
from numpy.typing import NDArray

from ropt.enums import VariableType
from ropt.evaluator import EvaluatorContext, EvaluatorResult
from ropt.plan import BasicOptimizer
from ropt.results import FunctionResults, Results

CONFIG: dict[str, Any] = {
    "variables": {
        "initial_values": 2 * [0.0],
        "lower_bounds": [0.0, 0.0],
        "upper_bounds": [10.0, 10.0],
        "types": VariableType.INTEGER,
    },
    "optimizer": {
        "method": "differential_evolution",
        "options": {"integrality": [True, True], "seed": 4},
        "max_functions": 100,
        "parallel": True,
    },
    "linear_constraints": {
        "coefficients": [1.0, 1.0],
        "lower_bounds": [-np.inf],
        "upper_bounds": [10.0],
    },
}


def function(variables: NDArray[np.float64], _: EvaluatorContext) -> EvaluatorResult:
    """Evaluate the function.

    Args:
        variables: The variables to evaluate.
        context:   Evaluator context.

    Returns:
        Calculated objectives.
    """
    x = variables[:, 0]
    y = variables[:, 1]
    objectives = -np.array(np.minimum(3 * x, y), ndmin=2).T
    return EvaluatorResult(objectives=objectives)


def report(results: tuple[Results, ...]) -> None:
    """Report results of an evaluation.

    Args:
        results: The results.
    """
    for item in results:
        if isinstance(item, FunctionResults) and item.functions is not None:
            print(f"  variables: {item.evaluations.variables}")
            print(f"  objective: {item.functions.weighted_objective}\n")


def run_optimization() -> None:
    """Run the optimization."""
    optimal_result = (
        BasicOptimizer(CONFIG, function).set_results_callback(report).run().results
    )
    assert optimal_result is not None
    assert optimal_result.functions is not None
    assert np.all(optimal_result.evaluations.variables == [3, 7])
    print(f"  variables: {optimal_result.evaluations.variables}")
    print(f"  objective: {optimal_result.functions.weighted_objective}\n")


def main() -> None:
    """Main function."""
    run_optimization()


if __name__ == "__main__":
    main()



"""The `EnOptConfig` configuration class."""

from __future__ import annotations

from typing import Any, Self

from pydantic import ConfigDict, ValidationInfo, model_validator

from ropt.config.utils import ImmutableBaseModel

from ._function_estimator_config import FunctionEstimatorConfig
from ._gradient_config import GradientConfig
from ._linear_constraints_config import LinearConstraintsConfig  # noqa: TC001
from ._nonlinear_constraints_config import NonlinearConstraintsConfig  # noqa: TC001
from ._objective_functions_config import ObjectiveFunctionsConfig
from ._optimizer_config import OptimizerConfig
from ._realization_filter_config import RealizationFilterConfig  # noqa: TC001
from ._realizations_config import RealizationsConfig
from ._sampler_config import SamplerConfig
from ._variables_config import VariablesConfig  # noqa: TC001


class EnOptConfig(ImmutableBaseModel):
    """The primary configuration class for an optimization step.

    `EnOptConfig` orchestrates the configuration of an entire optimization
    workflow. It contains nested configuration classes that define specific
    aspects of the optimization, such as variables, objectives, constraints,
    realizations, and the optimizer itself.

    `realization_filters`, `function_estimators`, and `samplers` are configured
    as tuples. Other configuration fields reference these objects by their index
    within the tuples. For example,
    [`GradientConfig`][ropt.config.enopt.GradientConfig] uses a `samplers`
    field, which is an array of indices specifying the sampler to use for each
    variable.

    Info:
        Many nested configuration classes use `numpy` arrays. These arrays
        typically have a size determined by a configured property (e.g., the
        number of variables) or a size of one. In the latter case, the single
        value is broadcasted to all relevant elements. For example,
        [`VariablesConfig`][ropt.config.enopt.VariablesConfig] defines
        properties like initial values and bounds as `numpy` arrays, which must
        either match the number of variables or have a size of one.

    Attributes:
        variables:             Configuration for the optimization variables.
        objectives:            Configuration for the objective functions.
        linear_constraints:    Configuration for linear constraints.
        nonlinear_constraints: Configuration for non-linear constraints.
        realizations:          Configuration for the realizations.
        optimizer:             Configuration for the optimization algorithm.
        gradient:              Configuration for gradient calculations.
        realization_filters:   Configuration for realization filters.
        function_estimators:   Configuration for function estimators.
        samplers:              Configuration for samplers.
    """

    variables: VariablesConfig
    objectives: ObjectiveFunctionsConfig = ObjectiveFunctionsConfig()
    linear_constraints: LinearConstraintsConfig | None = None
    nonlinear_constraints: NonlinearConstraintsConfig | None = None
    realizations: RealizationsConfig = RealizationsConfig()
    optimizer: OptimizerConfig = OptimizerConfig()
    gradient: GradientConfig = GradientConfig()
    realization_filters: tuple[RealizationFilterConfig, ...] = ()
    function_estimators: tuple[FunctionEstimatorConfig, ...] = (
        FunctionEstimatorConfig(),
    )
    samplers: tuple[SamplerConfig, ...] = (SamplerConfig(),)

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _linear_constraints(self, info: ValidationInfo) -> Self:
        if self.linear_constraints is not None:
            self._mutable()
            self.linear_constraints = self.linear_constraints.apply_transformation(
                self.variables, info.context
            )
            self._immutable()
        return self

    @model_validator(mode="after")
    def _gradient(self, info: ValidationInfo) -> Self:
        self._mutable()
        self.gradient = self.gradient.fix_perturbations(self.variables, info.context)
        self._immutable()
        return self

    @model_validator(mode="wrap")  # type: ignore[arg-type]
    def _pass_enopt_config_unchanged(self, handler: Any) -> Any:  # noqa: ANN401
        if isinstance(self, EnOptConfig):
            return self
        return handler(self)



"""This module implements the SciPy sampler plugin."""

import copy
import warnings
from typing import Any, Final

import numpy as np
from numpy.random import Generator
from numpy.typing import NDArray
from scipy.stats import norm, rv_continuous, truncnorm, uniform
from scipy.stats.qmc import Halton, LatinHypercube, QMCEngine, Sobol, scale

from ropt.config.enopt import EnOptConfig

from .base import Sampler, SamplerPlugin

_STATS_SAMPLERS: Final[dict[str, Any]] = {
    "uniform": uniform,
    "norm": norm,
    "truncnorm": truncnorm,
}

_QMC_ENGINES: Final = {
    "sobol": Sobol,
    "halton": Halton,
    "lhs": LatinHypercube,
}

_SUPPORTED_METHODS: Final[set[str]] = set(_STATS_SAMPLERS.keys()) | set(
    _QMC_ENGINES.keys()
)


class SciPySampler(Sampler):
    """Plugin class for producing sampling values via SciPy.

    This plugin implements the following sampling methods using the
    corresponding methods from the SciPy stats module:

    - Sampling from [probability
      distributions](https://docs.scipy.org/doc/scipy/reference/stats.html):

        `uniform`
        : Uniform distribution with a default range of [-1, 1].

        `norm`
        : Normal distribution with mean zero and standard deviation 1.

        `truncnorm`
        : Truncated normal distribution with mean zero and standard
          deviation 1 truncated a the range [-1, 1].

    - Sampling using methods from the [Quasi-Monte Carlo
      submodule](https://docs.scipy.org/doc/scipy/reference/stats.qmc.html):

        `sobol`
        : Using Sobol sequences, scaled to -1 and 1.

        `halton`
        : Using Halton sequences, scaled to -1 and 1.

        `lhs`
        : Using Latin Hypercube sampling, scaled to -1 and 1.

    Specific options that are normally passed as arguments in the SciPy
    functions can be provided via the options dictionary in the configuration
    object. Consult the
    [`scipy.stats`](https://docs.scipy.org/doc/scipy/reference/stats.html)
    manual for details on these options.
    """

    def __init__(
        self,
        enopt_config: EnOptConfig,
        sampler_index: int,
        mask: NDArray[np.bool_] | None,
        rng: Generator,
    ) -> None:
        """Initialize the sampler object.

        See the [ropt.plugins.sampler.base.Sample][] abstract base class.

        # noqa
        """
        self._enopt_config = enopt_config
        self._sampler_config = enopt_config.samplers[sampler_index]
        self._mask = mask
        _, _, self._method = self._sampler_config.method.lower().rpartition("/")
        if self._method == "default":
            self._method = "norm"
        self._rng = rng
        self._sampler: rv_continuous | QMCEngine
        self._options: dict[str, Any]
        if self._method not in _SUPPORTED_METHODS:
            msg = f"Method `{self._method}` is not implemented by the SciPy plugin"
            raise NotImplementedError(msg)
        self._sampler, self._options = self._init_sampler(self._sampler_config.options)

    def generate_samples(self) -> NDArray[np.float64]:
        """Generate a set of samples.

        See the [ropt.plugins.sampler.base.Sampler][] abstract base class.

        # noqa
        """
        variable_count = self._enopt_config.variables.initial_values.size
        realization_count = self._enopt_config.realizations.weights.size
        perturbation_count = self._enopt_config.gradient.number_of_perturbations

        sample_dim = variable_count if self._mask is None else self._mask.sum()

        if self._method in _STATS_SAMPLERS:
            samples = self._generate_stats_samples(
                1 if self._sampler_config.shared else realization_count,
                perturbation_count,
                sample_dim,
            )
        else:
            samples = self._generate_qmc_samples(
                1 if self._sampler_config.shared else realization_count,
                perturbation_count,
                sample_dim,
            )

        if self._sampler_config.shared:
            samples = np.repeat(samples, realization_count, axis=0)

        if self._mask is not None:
            shape = (realization_count, perturbation_count, variable_count)
            result = np.zeros(shape, dtype=np.float64)
            result[..., self._mask] = samples
            return result
        return samples

    def _init_sampler(
        self, options: dict[str, Any]
    ) -> tuple[rv_continuous | QMCEngine, dict[str, Any]]:
        options = copy.deepcopy(options)
        if self._method in _STATS_SAMPLERS:
            self._set_options(options)
            sampler = _STATS_SAMPLERS[self._method]
        elif self._method in _QMC_ENGINES:
            sample_dim = (
                self._enopt_config.variables.initial_values.size
                if self._mask is None
                else self._mask.sum()
            )
            sampler = _QMC_ENGINES[self._method](sample_dim, seed=self._rng, **options)
        else:
            msg = "sampler {self._method} is not supported by this SciPy version"
            raise NotImplementedError(msg)
        return sampler, options

    def _set_options(self, options: dict[str, Any]) -> None:
        parameters = {
            "uniform": {"loc": -1.0, "scale": 2.0},
            "truncnorm": {"a": -1.0, "b": 1.0},
        }
        if self._method in parameters:
            for key, value in parameters[self._method].items():
                options.setdefault(key, value)

    def _generate_stats_samples(
        self, realization_count: int, perturbation_count: int, sample_dim: int
    ) -> NDArray[np.float64]:
        return np.array(
            self._sampler.rvs(
                size=(realization_count, perturbation_count, sample_dim),
                random_state=self._rng,
                **self._options,
            ),
        )

    def _generate_qmc_samples(
        self, realization_count: int, perturbation_count: int, sample_dim: int
    ) -> NDArray[np.float64]:
        def _run_qmc_engine() -> NDArray[np.float64]:
            return np.array(
                scale(
                    self._sampler.random(realization_count * perturbation_count),
                    np.repeat(-1.0, sample_dim),
                    np.repeat(1.0, sample_dim),
                ).T.reshape((realization_count, perturbation_count, sample_dim)),
            )

        if self._method == "sobol":
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return _run_qmc_engine()
        else:
            return _run_qmc_engine()


class SciPySamplerPlugin(SamplerPlugin):
    """Default sampler plugin class."""

    @classmethod
    def create(
        cls,
        enopt_config: EnOptConfig,
        sampler_index: int,
        mask: NDArray[np.bool_] | None,
        rng: Generator,
    ) -> SciPySampler:
        """Initialize the sampler plugin.

        See the [ropt.plugins.sampler.base.SamplerPlugin][] abstract base class.

        # noqa
        """
        return SciPySampler(enopt_config, sampler_index, mask, rng)

    @classmethod
    def is_supported(cls, method: str) -> bool:
        """Check if a method is supported.

        See the [ropt.plugins.base.Plugin][] abstract base class.

        # noqa
        """
        return method.lower() in (_SUPPORTED_METHODS | {"default"})



from dataclasses import dataclass, field
from typing import Any, Protocol

import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig


@dataclass(slots=True)
class EvaluatorContext:
    """Capture additional details for the function evaluator.

    Function evaluator callbacks (see [`Evaluator`][ropt.evaluator.Evaluator])
    primarily receive variable vectors to evaluate objective and constraint
    functions. However, they may also benefit from additional information to
    optimize their calculations. This `EvaluatorContext` object provides that
    supplementary information.

    Specifically, it provides:

    - The configuration object for the current optimization step.
    - The realization index for each variable vector.
    - The perturbation index for each variable vector (if applicable). A value
      less than 0 indicates that the vector is not a perturbation.
    - Boolean matrices (`active_objectives` and `active_constraints`) indicating
      which objective/realization and constraint/realization evaluations are
      required by the optimizer.
    - A boolean vector (`active`) indicating which realizations require
      evaluation.

    The `active_objectives` and `active_constraints` matrices are structured
    such that each column corresponds to a realization, and each row corresponds
    to a function or constraint. A `True` value signifies that the corresponding
    evaluation is essential for the optimizer.

    Note: The `active` property
        In many cases, evaluators may only be able to compute all objectives and
        constraints for a given realization or none at all. In these scenarios,
        the `active` property provides a simplified view, indicating only the
        realizations that need to be evaluated. `active` cannot be set when creating
        the evaluator context, it is calculated from `active_objectives` and
        `active_constraints`.

    Args:
        config:             Configuration of the optimizer.
        realizations:       Realization numbers for each requested evaluation.
        perturbations:      Perturbation numbers for each requested evaluation.
                            A value less than 0 indicates that the vector is
                            not a perturbation.
        active_objectives:  Indicates which function/realization evaluations are
                            essential for the optimizer.
        active_constraints: Indicates which constraint/realization evaluations
                            are essential for the optimizer.
    """

    config: EnOptConfig
    realizations: NDArray[np.intc]
    perturbations: NDArray[np.intc] | None = None
    active_objectives: NDArray[np.bool_] | None = None
    active_constraints: NDArray[np.bool_] | None = None
    active: NDArray[np.bool_] | None = field(init=False)

    def __post_init__(self) -> None:
        active_objectives: NDArray[np.bool_] | None = (
            None
            if self.active_objectives is None
            else np.logical_or.reduce(self.active_objectives, axis=0)
        )
        active_constraints: NDArray[np.bool_] | None = (
            None
            if self.active_constraints is None
            else np.logical_or.reduce(self.active_constraints, axis=0)
        )
        if active_constraints is None:
            self.active = active_objectives
        elif active_objectives is None:
            self.active = active_constraints
        else:
            self.active = np.logical_or(active_objectives, active_constraints)


@dataclass
class EvaluatorResult:
    """Store the results of a function evaluation.

    This class stores the results of evaluating objective and constraint
    functions for a set of variable vectors.

    The `objectives` and `constraints` are stored as matrices. Each column in
    these matrices corresponds to a specific objective or constraint, and each
    row corresponds to a variable vector.

    When the evaluator is asked to evaluate functions, some variable vectors may
    be marked as inactive. The results for these inactive vectors should be set
    to zero. All active variable vectors should be evaluated. If an evaluation
    fails for any reason, the corresponding values should be set to `numpy.nan`.

    A `batch_id` can be set to identify this specific set of evaluation results.

    The `evaluation_info` dictionary can store additional metadata for each
    evaluation. This information is not used internally by `ropt` and can have
    an arbitrary structure, to be interpreted by the application. This can be
    used, for example, to uniquely identify the results calculated for each
    variable vector, allowing them to be linked back to their corresponding
    input vectors.

    Args:
        objectives:      The calculated objective values.
        constraints:     Optional calculated constraint values.
        batch_id:        Optional batch ID to identify this set of results.
        evaluation_info: Optional info for each evaluation.
    """

    objectives: NDArray[np.float64]
    constraints: NDArray[np.float64] | None = None
    batch_id: int | None = None
    evaluation_info: dict[str, NDArray[Any]] = field(default_factory=dict)


class Evaluator(Protocol):
    """Protocol for evaluator objects or callables.

    The [`EnsembleEvaluator`][ropt.ensemble_evaluator.EnsembleEvaluator] class
    requires a function evaluator callback that conforms to the
    [`Evaluator`][ropt.evaluator.Evaluator] signature. This callback accepts
    one or more variable vectors to evaluate, along with an
    [`EvaluatorContext`][ropt.evaluator.EvaluatorContext] object that provides
    relevant information for the evaluation. It returns an
    [`EvaluatorResult`][ropt.evaluator.EvaluatorResult] object containing the results.
    """

    def __call__(
        self, variables: NDArray[np.float64], context: EvaluatorContext, /
    ) -> EvaluatorResult:
        r"""Evaluate objective and constraint functions for given variables.

        This method defines the signature for the function evaluator callback.
        The evaluator calculates objective and constraint functions for a set of
        variable vectors, potentially for a subset of realizations and
        perturbations.

        Args:
            variables: The matrix of variables to evaluate. Each row represents
                       a variable vector.
            context:   The evaluation context, providing additional information
                       about the evaluation.

        Returns:
            An evaluation results object containing the calculated objective and
                constraint values, along with any additional metadata.

        Tip: Reusing Objective
            When defining multiple objectives, there may be a need to reuse the
            same objective value multiple times. For instance, a total objective
            could consist of the mean of the objectives for each realization,
            plus the standard deviation of the same values. This can be
            implemented by defining two objectives: the first calculated as the
            mean of the realizations, and the second using a function estimator
            to compute the standard deviations. The optimizer is unaware that
            both objectives use the same set of realizations. To prevent
            redundant calculations, the evaluator should compute the results of
            the realizations once and return them for both objectives.
        """



"""This module defines a basic optimization object."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Self

from ropt.config.enopt import EnOptConfig
from ropt.enums import EventType, OptimizerExitCode
from ropt.exceptions import OptimizationAborted

from ._context import OptimizerContext
from ._plan import Plan

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.evaluator import Evaluator
    from ropt.plan import Event
    from ropt.plugins.plan.base import ResultHandler
    from ropt.results import FunctionResults
    from ropt.transforms import OptModelTransforms


@dataclass(slots=True)
class _Results:
    results: FunctionResults | None
    variables: NDArray[np.float64] | None
    exit_code: OptimizerExitCode = OptimizerExitCode.UNKNOWN


class BasicOptimizer:
    """A class for executing single optimization runs.

    The `BasicOptimizer` is designed to simplify the process of setting up and
    executing optimization workflows that consist primarily of a single
    optimization run. It offers a more streamlined approach compared to directly
    defining and managing a full `Plan` object, making it ideal for
    straightforward optimization tasks.

    This class provides a user-friendly interface for common optimization
    operations, including:

    - **Initiating a Single Optimization:**  Easily start an optimization
      process with a provided configuration and evaluator.
    - **Observing Optimization Events:** Register observer functions to monitor
      and react to various events that occur during the optimization, such as
      the start of an evaluation or the availability of new results.
    - **Abort Conditions:** Define a callback function that can be used to check
      for abort conditions during the optimization.
    - **Result Reporting:** Define a callback function that will be called
      whenever new results become available.
    - **Accessing Results:** After the optimization is complete, the optimal
      results, corresponding variables, and the optimization's exit code are
      readily accessible.
    - **Customizable Steps and Handlers:** While designed for single runs, it
      allows for the addition of custom steps and handlers to the underlying
      `Plan` for more complex scenarios. It is possible to pass keyword
      arguments to the custom steps and handlers.

    By encapsulating the core elements of an optimization run, the
    `BasicOptimizer` reduces the boilerplate code required for simple
    optimization tasks, allowing users to focus on defining the optimization
    problem and analyzing the results.
    """

    def __init__(
        self,
        enopt_config: dict[str, Any] | EnOptConfig,
        evaluator: Evaluator,
        *,
        transforms: OptModelTransforms | None = None,
        constraint_tolerance: float = 1e-10,
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize a `BasicOptimizer` object.

        This constructor sets up the necessary components for a single
        optimization run. It requires an optimization configuration and an
        evaluator, which together define the optimization problem and how to
        evaluate potential solutions. The `transforms` object can be used to
        apply transformations to the optimization model, such as scaling or
        shifting variables. If a constraint value is within the
        `constraint_tolerance` of zero, it is considered satisfied. The `kwargs`
        may be used to define custom steps, and handlers to modify the behavior
        of the optimization process.

        Note: Custom  steps
            The optional keyword arguments (`kwargs`) provide a mechanism to
            inject a custom step into the optimization process. The behavior is
            as follows:

            1.  **Custom Step Execution:** If a single keyword argument is
                provided, the `BasicOptimizer` checks if a step with the same
                name exists. If so, that step is executed immediately, receiving
                the key-value pair as a keyword input, together with the
                transforms passed via a `transforms` keyword. Only one custom
                step can be executed this way, if other keyword arguments are
                present an error is raised. The custom step receives the `Plan`
                object and may install a custom run function to be executed
                later, or install custom result handlers.
            2.  **Default Optimization:** If no custom step is run, or if the
                custom step does not install a custom run function, the default
                optimization process is used.
            3.  **Callback Installation and Execution:** Finally, any callbacks
                added via `set_abort_callback` or `set_results_callback` are
                installed, and the appropriate run function is executed.

        Args:
            enopt_config:         The configuration for the optimization.
            evaluator:            The evaluator object.
            transforms:           Optional transforms.
            constraint_tolerance: The constraint violation tolerance.
            kwargs:               Optional keyword arguments.
        """
        self._config = EnOptConfig.model_validate(enopt_config)
        self._transforms = transforms
        self._constraint_tolerance = constraint_tolerance
        self._optimizer_context = OptimizerContext(evaluator=evaluator)
        self._observers: list[tuple[EventType, Callable[[Event], None]]] = []
        self._results: _Results
        self._kwargs: dict[str, Any] = kwargs

    @property
    def results(self) -> FunctionResults | None:
        """Return the optimal result found during the optimization.

        This property provides access to the best
        [`FunctionResults`][ropt.results.FunctionResults] object discovered
        during the optimization process. It encapsulates the objective function
        value, constraint values, and other relevant information about the
        optimal solution.

        Returns:
            The optimal result.
        """
        return self._results.results

    @property
    def variables(self) -> NDArray[np.float64] | None:
        """Return the optimal variables found during the optimization.

        This property provides access to the variable values that correspond to
        the optimal [`FunctionResults`][ropt.results.FunctionResults] object.
        These variables represent the solution that yielded the best objective
        function value found during the optimization process.

        Returns:
            The variables corresponding to the optimal result.
        """
        return self._results.variables

    @property
    def exit_code(self) -> OptimizerExitCode:
        """Return the exit code of the optimization run.

        This property provides access to the
        [`OptimizerExitCode`][ropt.enums.OptimizerExitCode] that indicates the
        outcome of the optimization process. It can be used to determine whether
        the optimization completed successfully, was aborted, or encountered an
        error.

        Returns:
            The exit code of the optimization run.
        """
        return self._results.exit_code

    def run(self) -> Self:
        """Run the optimization process.

        This method initiates the optimization workflow defined by the
        `BasicOptimizer` object. It executes the underlying `Plan`, which
        manages the optimization steps, result handling, and event processing.
        After the optimization is complete, the optimal results, variables, and
        exit code can be accessed via the corresponding properties.

        Returns:
            The `BasicOptimizer` instance, allowing for method chaining.
        """

        def _run_func(
            plan: Plan, transforms: OptModelTransforms | None
        ) -> tuple[ResultHandler | None, OptimizerExitCode | None]:
            exit_code = plan.run_step(
                optimizer, config=self._config, transforms=transforms
            )
            return plan.get(tracker, "results"), exit_code

        plan = Plan(self._optimizer_context)

        # Optionally run a custom step defined in the keyword arguments:
        key, value = next(iter(self._kwargs.items()), (None, None))
        if key is not None and plan.step_exists(key):
            if len(self._kwargs) > 1:
                msg = "Only one custom step is allowed."
                raise TypeError(msg)
            plan.run_step(
                plan.add_step(key), transforms=self._transforms, **{key: value}
            )

        # If no custom function was installed, install the default function:
        if not plan.has_function():
            optimizer = plan.add_step("optimizer")
            tracker = plan.add_handler(
                "tracker",
                constraint_tolerance=self._constraint_tolerance,
                sources={optimizer},
                transforms=self._transforms,
            )
            plan.add_function(_run_func)

        for event_type, function in self._observers:
            self._optimizer_context.add_observer(event_type, function)

        results, exit_code = plan.run_function(self._transforms)
        variables = None if results is None else results.evaluations.variables
        self._results = _Results(
            results=results,
            variables=variables,
            exit_code=exit_code,
        )
        return self

    def set_abort_callback(self, callback: Callable[[], bool]) -> Self:
        """Set a callback to check for abort conditions.

        The provided callback function will be invoked repeatedly during the
        optimization process. If the callback returns `True`, the optimization
        will be aborted, and the `BasicOptimizer` will exit with an
        [`OptimizerExitCode.USER_ABORT`][ropt.enums.OptimizerExitCode].

        The callback function should have no arguments and return a boolean
        value.

        Args:
            callback: The callable to check for abort conditions.

        Returns:
            The `BasicOptimizer` instance, allowing for method chaining.
        """

        def _check_abort_callback(_: Event) -> None:
            if callback():
                raise OptimizationAborted(exit_code=OptimizerExitCode.USER_ABORT)

        self._observers.append((EventType.START_EVALUATION, _check_abort_callback))
        return self

    def set_results_callback(
        self,
        callback: Callable[..., None],
    ) -> Self:
        """Set a callback to report new results.

        The provided callback function will be invoked whenever new results
        become available during the optimization process. This allows for
        real-time monitoring and analysis of the optimization's progress.

        The required signature of the callback function should be:

        ```python
        def callback(results: tuple[FunctionResults, ...]) -> None:
            ...
        ```

        Args:
            callback: The callable that will be invoked to report new results.

        Returns:
            The `BasicOptimizer` instance, allowing for method chaining.
        """

        def _results_callback(event: Event) -> None:
            results = event.data.get("results", ())
            if self._transforms is not None:
                results = tuple(
                    item.transform_from_optimizer(self._transforms) for item in results
                )
            callback(results)

        self._observers.append((EventType.FINISHED_EVALUATION, _results_callback))
        return self



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any

import numpy as np
import pytest

from ropt.enums import EventType, OptimizerExitCode
from ropt.exceptions import OptimizationAborted, PlanAborted
from ropt.plan import (
    BasicOptimizer,
    Event,
    OptimizerContext,
    Plan,
)
from ropt.results import FunctionResults

if TYPE_CHECKING:
    from numpy.typing import NDArray

# ruff: noqa: SLF001


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> dict[str, Any]:
    return {
        "optimizer": {
            "tolerance": 1e-5,
            "max_functions": 20,
        },
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "objectives": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def test_run_basic(enopt_config: dict[str, Any], evaluator: Any) -> None:
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker = plan.add_handler("tracker", sources={step})
    plan.run_step(step, config=enopt_config)
    assert plan.get(tracker, "results") is not None
    assert np.allclose(
        plan.get(tracker, "results").evaluations.variables, [0.0, 0.0, 0.5], atol=0.02
    )

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)


def test_plan_rng(enopt_config: dict[str, Any], evaluator: Any) -> None:
    enopt_config["gradient"]["seed"] = 1
    config2 = deepcopy(enopt_config)
    config2["gradient"]["seed"] = 2

    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker = plan.add_handler("tracker", sources={step})

    plan.run_step(step, config=enopt_config)
    assert plan.get(tracker, "results") is not None
    variables1 = plan.get(tracker, "results").evaluations.variables
    plan.set(tracker, "results", None)
    plan.run_step(step, config=enopt_config)
    assert plan.get(tracker, "results") is not None
    variables2 = plan.get(tracker, "results").evaluations.variables
    plan.set(tracker, "results", None)
    plan.run_step(step, config=config2)
    assert plan.get(tracker, "results") is not None
    variables3 = plan.get(tracker, "results").evaluations.variables

    assert np.all(variables1 == variables2)
    assert not np.all(variables2 == variables3)


def test_set_initial_values(enopt_config: dict[str, Any], evaluator: Any) -> None:
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker = plan.add_handler("tracker", sources={step})
    plan.run_step(step, config=enopt_config)
    assert plan.get(tracker, "results") is not None
    variables1 = plan.get(tracker, "results").evaluations.variables
    plan.set(tracker, "variables", None)
    plan.run_step(step, config=enopt_config, variables=[0, 0, 0])
    assert plan.get(tracker, "results") is not None
    variables2 = plan.get(tracker, "results").evaluations.variables

    assert variables1 is not None
    assert variables2 is not None

    assert np.allclose(variables1, [0.0, 0.0, 0.5], atol=0.02)
    assert np.allclose(variables2, [0.0, 0.0, 0.5], atol=0.02)

    assert not np.all(variables1 == variables2)


def test_reset_results(enopt_config: dict[str, Any], evaluator: Any) -> None:
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker = plan.add_handler("tracker", sources={step})
    plan.run_step(step, config=enopt_config)
    saved_results = deepcopy(plan.get(tracker, "results"))
    plan.set(tracker, "results", None)
    assert saved_results is not None
    assert np.allclose(saved_results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)


def test_two_optimizers_alternating(
    enopt_config: dict[str, Any], evaluator: Any
) -> None:
    completed_functions = 0

    def _track_evaluations(event: Event) -> None:
        nonlocal completed_functions

        for item in event.data["results"]:
            if isinstance(item, FunctionResults):
                completed_functions += 1

    enopt_config["variables"]["initial_values"] = [0.0, 0.2, 0.1]

    opt_config1 = {
        "speculative": True,
        "max_functions": 4,
    }
    opt_config2 = {
        "speculative": True,
        "max_functions": 3,
    }

    enopt_config1 = deepcopy(enopt_config)
    enopt_config1["variables"]["mask"] = [True, False, True]
    enopt_config1["optimizer"] = opt_config1
    enopt_config2 = deepcopy(enopt_config)
    enopt_config2["variables"]["mask"] = [False, True, False]
    enopt_config2["optimizer"] = opt_config2

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker1 = plan.add_handler("tracker", sources={step})
    tracker2 = plan.add_handler("tracker", sources={step}, what="last")
    plan.run_step(step, config=enopt_config1)
    assert plan.get(tracker2, "results") is not None
    plan.run_step(
        step,
        config=enopt_config2,
        variables=plan.get(tracker2, "results").evaluations.variables,
    )
    assert plan.get(tracker2, "results") is not None
    plan.run_step(
        step,
        config=enopt_config1,
        variables=plan.get(tracker2, "results").evaluations.variables,
    )
    assert plan.get(tracker2, "results") is not None
    plan.run_step(
        step,
        config=enopt_config2,
        variables=plan.get(tracker2, "results").evaluations.variables,
    )
    assert completed_functions == 14
    assert plan.get(tracker1, "results") is not None
    assert np.allclose(
        plan.get(tracker1, "results").evaluations.variables, [0.0, 0.0, 0.5], atol=0.02
    )


def test_optimization_sequential(enopt_config: dict[str, Any], evaluator: Any) -> None:
    completed: list[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed

        completed += [
            item for item in event.data["results"] if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 2

    enopt_config2 = deepcopy(enopt_config)
    enopt_config2["optimizer"]["max_functions"] = 3

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker = plan.add_handler("tracker", sources={step}, what="last")
    plan.run_step(step, config=enopt_config)
    assert plan.get(tracker, "results") is not None
    plan.run_step(
        step,
        config=enopt_config2,
        variables=plan.get(tracker, "results").evaluations.variables,
    )

    assert not np.allclose(
        completed[1].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02
    )
    assert np.all(
        completed[2].evaluations.variables == completed[1].evaluations.variables
    )
    assert np.allclose(completed[-1].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)


def test_restart_initial(enopt_config: dict[str, Any], evaluator: Any) -> None:
    completed: list[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed

        completed += [
            item for item in event.data["results"] if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 3

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(context)
    step = plan.add_step("optimizer")
    for _ in range(2):
        plan.run_step(step, config=enopt_config)

    assert len(completed) == 6
    initial = np.array([enopt_config["variables"]["initial_values"]])
    assert np.all(completed[0].evaluations.variables == initial)
    assert np.all(completed[3].evaluations.variables == initial)


def test_restart_last(enopt_config: dict[str, Any], evaluator: Any) -> None:
    completed: list[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed

        completed += [
            item for item in event.data["results"] if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 3

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker = plan.add_handler("tracker", sources={step}, what="last")
    for _ in range(2):
        variables = (
            None
            if plan.get(tracker, "results") is None
            else plan.get(tracker, "results").evaluations.variables
        )
        plan.run_step(step, config=enopt_config, variables=variables)

    assert np.all(
        completed[3].evaluations.variables == completed[2].evaluations.variables
    )


def test_restart_optimum(enopt_config: dict[str, Any], evaluator: Any) -> None:
    completed: list[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed

        completed += [
            item for item in event.data["results"] if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker = plan.add_handler("tracker", sources={step})
    for _ in range(2):
        variables = (
            None
            if plan.get(tracker, "results") is None
            else plan.get(tracker, "results").evaluations.variables
        )
        plan.run_step(step, config=enopt_config, variables=variables)

    assert np.all(
        completed[2].evaluations.variables == completed[4].evaluations.variables
    )


def test_restart_optimum_with_reset(
    enopt_config: dict[str, Any], evaluator: Any, test_functions: Any
) -> None:
    completed: list[FunctionResults] = []
    max_functions = 5

    def _track_evaluations(event: Event) -> None:
        nonlocal completed

        completed += [
            item for item in event.data["results"] if isinstance(item, FunctionResults)
        ]

    # Make sure each restart has worse objectives, and that the last evaluation
    # is even worse, so each run has its own optimum that is worse than the
    # global and not at its last evaluation. We should end up with initial
    # values that are not from the global optimum, or from the most recent
    # evaluation:
    new_functions = (
        lambda variables, context: (
            test_functions[0](variables, context)
            + int((len(completed) + 1) / max_functions)
        ),
        lambda variables, context: (
            test_functions[1](variables, context)
            + int((len(completed) + 1) / max_functions)
        ),
    )

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = max_functions

    context = OptimizerContext(evaluator=evaluator(new_functions)).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker = plan.add_handler("tracker", sources={step})
    for _ in range(3):
        variables = (
            None
            if plan.get(tracker, "results") is None
            else plan.get(tracker, "results").evaluations.variables
        )
        plan.set(tracker, "results", None)
        plan.run_step(step, config=enopt_config, variables=variables)

    # The third evaluation is the optimum, and used to restart the second run:
    assert np.all(
        completed[max_functions].evaluations.variables
        == completed[2].evaluations.variables
    )
    # The 5th evaluation is the optimum of the second run, and used for the third:
    assert np.all(
        completed[2 * max_functions].evaluations.variables
        == completed[5].evaluations.variables
    )


def test_repeat_metadata(enopt_config: dict[str, Any], evaluator: Any) -> None:
    restarts: list[int] = []

    def _track_results(event: Event) -> None:
        metadata = event.data["results"][0].metadata
        restart = metadata.get("restart", -1)
        assert metadata["foo"] == 1
        assert metadata["bar"] == "string"
        if not restarts or restart != restarts[-1]:
            restarts.append(restart)

    metadata = {
        "foo": 1,
        "bar": "string",
    }

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_results
    )
    plan = Plan(context)
    step = plan.add_step("optimizer")
    for idx in range(2):
        metadata["restart"] = idx
        plan.run_step(step, config=enopt_config, metadata=metadata)
    assert restarts == [0, 1]


def test_evaluator_step(enopt_config: dict[str, Any], evaluator: Any) -> None:
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(context)
    step = plan.add_step("evaluator")
    tracker = plan.add_handler("tracker", sources={step})
    plan.run_step(step, config=enopt_config)
    assert plan.get(tracker, "results").functions is not None
    assert np.allclose(plan.get(tracker, "results").functions.weighted_objective, 1.66)

    plan.set(tracker, "results", None)
    plan.run_step(step, config=enopt_config, variables=[0, 0, 0])
    assert plan.get(tracker, "results").functions is not None
    assert np.allclose(plan.get(tracker, "results").functions.weighted_objective, 1.75)


def test_evaluator_step_multi(enopt_config: dict[str, Any], evaluator: Any) -> None:
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(context)
    step = plan.add_step("evaluator")
    store = plan.add_handler("store", sources={step})
    plan.run_step(step, config=enopt_config, variables=[[0, 0, 0.1], [0, 0, 0]])
    values = [
        results.functions.weighted_objective.item()
        for results in plan.get(store, "results")
    ]
    assert np.allclose(values, [1.66, 1.75])


def test_exit_code(enopt_config: dict[str, Any], evaluator: Any) -> None:
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(context)
    step = plan.add_step("optimizer")
    exit_code = plan.run_step(step, config=enopt_config)
    assert exit_code == OptimizerExitCode.MAX_FUNCTIONS_REACHED


def test_nested_plan(enopt_config: dict[str, Any], evaluator: Any) -> None:
    enopt_config["variables"]["initial_values"] = [0.0, 0.2, 0.1]

    completed_functions = 0

    def _track_evaluations(event: Event) -> None:
        nonlocal completed_functions

        for item in event.data["results"]:
            if isinstance(item, FunctionResults):
                completed_functions += 1

    enopt_config["optimizer"]["tolerance"] = 1e-10
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4
    enopt_config["variables"]["mask"] = [True, False, True]
    nested_config = deepcopy(enopt_config)
    nested_config["variables"]["mask"] = [False, True, False]
    enopt_config["optimizer"]["max_functions"] = 5

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )

    inner_plan = Plan(context)
    inner_step = inner_plan.add_step("optimizer")
    inner_tracker = inner_plan.add_handler("tracker", sources={inner_step})

    def _inner_optimization(
        plan: Plan, variables: NDArray[np.float64]
    ) -> FunctionResults | None:
        plan.run_step(inner_step, config=nested_config, variables=variables)
        results = inner_plan.get(inner_tracker, "results")
        assert isinstance(results, FunctionResults | None)
        return results

    inner_plan.add_function(_inner_optimization)

    outer_plan = Plan(context)
    outer_step = outer_plan.add_step("optimizer")
    outer_tracker = outer_plan.add_handler("tracker", sources={outer_step})
    outer_plan.run_step(outer_step, config=enopt_config, nested_optimization=inner_plan)
    assert outer_plan.get(outer_tracker, "results") is not None
    assert np.allclose(
        outer_plan.get(outer_tracker, "results").evaluations.variables,
        [0.0, 0.0, 0.5],
        atol=0.02,
    )
    assert completed_functions == 25


def test_nested_plan_metadata(enopt_config: dict[str, Any], evaluator: Any) -> None:
    enopt_config["variables"]["initial_values"] = [0.0, 0.2, 0.1]

    def _track_evaluations(event: Event) -> None:
        for item in event.data["results"]:
            if isinstance(item, FunctionResults):
                if event.source == "outer":
                    assert item.metadata.get("outer") == 1
                if event.source == "inner":
                    assert item.metadata.get("inner") == "inner_meta_data"

    enopt_config["optimizer"]["tolerance"] = 1e-10
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4
    enopt_config["variables"]["mask"] = [True, False, True]
    nested_config = deepcopy(enopt_config)
    nested_config["variables"]["mask"] = [False, True, False]
    enopt_config["optimizer"]["max_functions"] = 5

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )

    inner_plan = Plan(context)
    inner_step = inner_plan.add_step("optimizer")
    inner_tracker = inner_plan.add_handler("tracker", sources={inner_step})

    def _inner_optimization(
        plan: Plan, variables: NDArray[np.float64]
    ) -> FunctionResults | None:
        plan.run_step(
            inner_step,
            config=nested_config,
            metadata={"inner": "inner_meta_data"},
            variables=variables,
        )
        results = inner_plan.get(inner_tracker, "results")
        assert isinstance(results, FunctionResults)
        return results

    inner_plan.add_function(_inner_optimization)

    outer_plan = Plan(context)
    outer_step = outer_plan.add_step("optimizer")
    outer_tracker = outer_plan.add_handler("tracker", sources={outer_step})
    outer_plan.run_step(
        outer_step,
        config=enopt_config,
        nested_optimization=inner_plan,
        metadata={"outer": 1},
    )

    assert inner_plan.get(inner_tracker, "results") is not None
    assert np.allclose(
        outer_plan.get(outer_tracker, "results").evaluations.variables,
        [0.0, 0.0, 0.5],
        atol=0.02,
    )
    assert outer_plan.get(outer_tracker, "results").metadata["outer"] == 1
    assert (
        inner_plan.get(inner_tracker, "results").metadata["inner"] == "inner_meta_data"
    )


def test_plan_abort(enopt_config: Any, evaluator: Any) -> None:
    last_evaluation = 0

    def _observer(_: Event) -> None:
        nonlocal last_evaluation

        last_evaluation += 1
        if last_evaluation == 1:
            raise OptimizationAborted(exit_code=OptimizerExitCode.USER_ABORT)

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _observer
    )
    plan = Plan(context)
    step = plan.add_step("optimizer")
    tracker = plan.add_handler("tracker", sources={step})
    plan.run_step(step, config=enopt_config)
    assert plan.get(tracker, "results") is not None
    assert last_evaluation == 1
    assert plan.aborted

    step = plan.add_step("optimizer")
    with pytest.raises(PlanAborted, match="Plan was aborted by the previous step."):
        plan.run_step(step, config=enopt_config)



"""Configuration class for gradients."""

from __future__ import annotations

from typing import TYPE_CHECKING, Self

import numpy as np
from pydantic import ConfigDict, PositiveInt, model_validator

from ropt.config.utils import ImmutableBaseModel, check_enum_values, immutable_array
from ropt.config.validated_types import (  # noqa: TC001
    Array1D,
    Array1DInt,
    ArrayEnum,
    ItemOrTuple,
)
from ropt.enums import BoundaryType, PerturbationType

from .constants import (
    DEFAULT_NUMBER_OF_PERTURBATIONS,
    DEFAULT_PERTURBATION_BOUNDARY_TYPE,
    DEFAULT_PERTURBATION_MAGNITUDE,
    DEFAULT_PERTURBATION_TYPE,
    DEFAULT_SEED,
)

if TYPE_CHECKING:
    from ropt.config.enopt import VariablesConfig
    from ropt.transforms import OptModelTransforms


class GradientConfig(ImmutableBaseModel):
    """Configuration class for gradient calculations.

    This class, `GradientConfig`, defines the configuration for gradient
    calculations used in an [`EnOptConfig`][ropt.config.enopt.EnOptConfig]
    object.

    Gradients are estimated using function values calculated from perturbed
    variables and the unperturbed variables. The `number_of_perturbations` field
    determines the number of perturbed variables used, which must be at least
    one.

    If function evaluations for some perturbed variables fail, the gradient may
    still be estimated as long as a minimum number of evaluations succeed. The
    `perturbation_min_success` field specifies this minimum. By default, it
    equals `number_of_perturbations`.

    Perturbations are generated by sampler objects configured in the parent
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object. The `samplers` field
    specifies, for each variable, the index of the sampler to use. A random
    number generator is created to support samplers that require random numbers.

    The generated perturbation values are scaled by the
    `perturbation_magnitudes` and can be modified based on the
    `perturbation_types`. See [`PerturbationType`][ropt.enums.PerturbationType]
    for details on available perturbation types.

    Perturbed variables may violate the defined variable bounds. The
    `boundary_types` field specifies how to handle such violations. See
    [`BoundaryType`][ropt.enums.BoundaryType] for details on available boundary
    handling methods.

    The `perturbation_types` and `boundary_types` fields use values from the
    [`PerturbationType`][ropt.enums.PerturbationType] and
    [`BoundaryType`][ropt.enums.BoundaryType] enumerations, respectively.

    Gradients are calculated for each realization individually and then combined
    into a total gradient. If `number_of_perturbations` is low, or even just
    one, individual gradient calculations may be unreliable. In this case,
    setting `merge_realizations` to `True` directs the optimizer to combine the
    results of all realizations directly into a single gradient estimate.

    Note: Seed for Samplers
        The `seed` value ensures consistent results across repeated runs with
        the same configuration. To obtain unique results for each optimization
        run, modify the seed. A common approach is to use a tuple with a unique
        ID as the first element, ensuring reproducibility across nested and
        parallel plan evaluations.

    Attributes:
        number_of_perturbations:  Number of perturbations (default:
            [`DEFAULT_NUMBER_OF_PERTURBATIONS`][ropt.config.enopt.constants.DEFAULT_NUMBER_OF_PERTURBATIONS]).
        perturbation_min_success: Minimum number of successful function evaluations
            for perturbed variables (default: equal to `number_of_perturbations`).
        perturbation_magnitudes:  Magnitudes of the perturbations for each variable
            (default:
            [`DEFAULT_PERTURBATION_MAGNITUDE`][ropt.config.enopt.constants.DEFAULT_PERTURBATION_MAGNITUDE]).
        perturbation_types:       Type of perturbation for each variable (see
            [`PerturbationType`][ropt.enums.PerturbationType], default:
            [`DEFAULT_PERTURBATION_TYPE`][ropt.config.enopt.constants.DEFAULT_PERTURBATION_TYPE]).
        boundary_types:           How to handle perturbations that violate boundary
            conditions (see [`BoundaryType`][ropt.enums.BoundaryType], default:
            [`DEFAULT_PERTURBATION_BOUNDARY_TYPE`][ropt.config.enopt.constants.DEFAULT_PERTURBATION_BOUNDARY_TYPE]).
        samplers:                 Indices of the samplers to use for each variable.
        seed:                     Seed for the random number generator used by the samplers.
        merge_realizations:       Merge all realizations for the final gradient
            calculation (default: `False`).
    """

    number_of_perturbations: PositiveInt = DEFAULT_NUMBER_OF_PERTURBATIONS
    perturbation_min_success: PositiveInt | None = None
    perturbation_magnitudes: Array1D = np.array(DEFAULT_PERTURBATION_MAGNITUDE)
    perturbation_types: ArrayEnum = np.array(DEFAULT_PERTURBATION_TYPE)
    boundary_types: ArrayEnum = np.array(DEFAULT_PERTURBATION_BOUNDARY_TYPE)
    samplers: Array1DInt | None = None
    seed: ItemOrTuple[int] = (DEFAULT_SEED,)
    merge_realizations: bool = False

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _check_perturbation_min_success(self) -> Self:
        self._mutable()
        if (
            self.perturbation_min_success is None
            or self.perturbation_min_success > self.number_of_perturbations
        ):
            self.perturbation_min_success = self.number_of_perturbations
        self._immutable()
        return self

    @model_validator(mode="after")
    def _check(self) -> Self:
        check_enum_values(self.perturbation_types, PerturbationType)
        check_enum_values(self.boundary_types, BoundaryType)
        return self

    def fix_perturbations(
        self, variables: VariablesConfig, transforms: OptModelTransforms | None
    ) -> GradientConfig:
        variable_count = variables.initial_values.size
        magnitudes = self.perturbation_magnitudes
        boundary_types = self.boundary_types
        types = self.perturbation_types

        try:
            magnitudes = np.broadcast_to(magnitudes, (variable_count,))
        except ValueError as err:
            msg = (
                "the perturbation magnitudes cannot be broadcasted "
                f"to a length of {variable_count}"
            )
            raise ValueError(msg) from err

        if boundary_types.size == 1:
            boundary_types = np.broadcast_to(
                immutable_array(boundary_types),
                (variable_count,),
            )
        elif boundary_types.size == variable_count:
            boundary_types = immutable_array(boundary_types)
        else:
            msg = f"perturbation boundary_types must have {variable_count} items"
            raise ValueError(msg)

        if types.size == 1:
            types = np.broadcast_to(immutable_array(types), (variable_count,))
        elif types.size == variable_count:
            types = immutable_array(types)
        else:
            msg = f"perturbation types must have {variable_count} items"
            raise ValueError(msg)

        relative = types == PerturbationType.RELATIVE
        if not np.all(
            np.logical_and(
                np.isfinite(variables.lower_bounds[relative]),
                np.isfinite(variables.upper_bounds[relative]),
            ),
        ):
            msg = "The variable bounds must be finite to use relative perturbations"
            raise ValueError(msg)
        magnitudes = np.where(
            relative,
            (variables.upper_bounds - variables.lower_bounds) * magnitudes,
            magnitudes,
        )

        if transforms is not None and transforms.variables is not None:
            absolute = types == PerturbationType.ABSOLUTE
            transformed = transforms.variables.magnitudes_to_optimizer(magnitudes)
            magnitudes[absolute] = transformed[absolute]

        return self.model_copy(
            update={
                "perturbation_magnitudes": magnitudes,
                "boundary_types": boundary_types,
                "perturbation_types": types,
            }
        )
